from copy import deepcopy

import numpy as np
import ocelot
import torch

import cheetah

from .resources import ARESlatticeStage3v1_9 as ares


def test_dipole():
    """
    Test that the tracking results through a Cheeath `Dipole` element match those
    through an Oclet `Bend` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_dipole = cheetah.Dipole(length=torch.tensor(0.1), angle=torch.tensor(0.1))
    outgoing_beam = cheetah_dipole.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_bend = ocelot.Bend(l=0.1, angle=0.1)
    lattice = ocelot.MagneticLattice([ocelot_bend])
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert np.allclose(
        outgoing_beam.particles[:, :6], outgoing_p_array.rparticles.transpose()
    )


def test_dipole_with_fringe_field():
    """
    Test that the tracking results through a Cheeath `Dipole` element match those
    through an Oclet `Bend` element when there are fringe fields.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_dipole = cheetah.Dipole(
        length=torch.tensor(0.1),
        angle=torch.tensor(0.1),
        fringe_integral=torch.tensor(0.1),
        gap=torch.tensor(0.2),
    )
    outgoing_beam = cheetah_dipole.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_bend = ocelot.Bend(l=0.1, angle=0.1, fint=0.1, gap=0.2)
    lattice = ocelot.MagneticLattice([ocelot_bend])
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert np.allclose(
        outgoing_beam.particles[:, :6], outgoing_p_array.rparticles.transpose()
    )


def test_aperture():
    """
    Test that the tracking results through a Cheeath `Aperture` element match those
    through an Oclet `Aperture` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Aperture(
                x_max=torch.tensor(2e-4),
                y_max=torch.tensor(2e-4),
                shape="rectangular",
                name="aperture",
                is_active=True,
            ),
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_cell = [ocelot.Aperture(xmax=2e-4, ymax=2e-4), ocelot.Drift(l=0.1)]
    lattice = ocelot.MagneticLattice([ocelot_cell])
    navigator = ocelot.Navigator(lattice)
    navigator.activate_apertures()
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert outgoing_beam.num_particles == outgoing_p_array.rparticles.shape[1]


def test_aperture_elliptical():
    """
    Test that the tracking results through an elliptical Cheeath `Aperture` element
    match those through an elliptical Oclet `Aperture` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Aperture(
                x_max=torch.tensor(2e-4),
                y_max=torch.tensor(2e-4),
                shape="elliptical",
                name="aperture",
                is_active=True,
            ),
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_cell = [
        ocelot.Aperture(xmax=2e-4, ymax=2e-4, type="ellipt"),
        ocelot.Drift(l=0.1),
    ]
    lattice = ocelot.MagneticLattice([ocelot_cell])
    navigator = ocelot.Navigator(lattice)
    navigator.activate_apertures()
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert outgoing_beam.num_particles == outgoing_p_array.rparticles.shape[1]


def test_solenoid():
    """
    Test that the tracking results through a Cheeath `Solenoid` element match those
    through an Oclet `Solenoid` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_solenoid = cheetah.Solenoid(length=torch.tensor(0.5), k=torch.tensor(5.0))
    outgoing_beam = cheetah_solenoid.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_solenoid = ocelot.Solenoid(l=0.5, k=5.0)
    lattice = ocelot.MagneticLattice([ocelot_solenoid])
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert np.allclose(
        outgoing_beam.particles[:, :6], outgoing_p_array.rparticles.transpose()
    )


def test_ares_ea():
    """
    Test that the tracking results through a Experimental Area (EA) lattice of the ARES
    accelerator at DESY match those using Ocelot.
    """
    cell = cheetah.converters.nocelot.subcell_of_ocelot(
        ares.cell, "AREASOLA1", "AREABSCR1"
    )
    ares.areamqzm1.k1 = 5.0
    ares.areamqzm2.k1 = -5.0
    ares.areamcvm1.k1 = 1e-3
    ares.areamqzm3.k1 = 5.0
    ares.areamchm1.k1 = -2e-3

    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_segment = cheetah.Segment.from_ocelot(cell)
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    lattice = ocelot.MagneticLattice(cell)
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    assert np.isclose(outgoing_beam.mu_x, outgoing_p_array.x().mean())
    assert np.isclose(outgoing_beam.mu_xp, outgoing_p_array.px().mean())
    assert np.isclose(outgoing_beam.mu_y, outgoing_p_array.y().mean())
    assert np.isclose(outgoing_beam.mu_yp, outgoing_p_array.py().mean())
    assert np.isclose(outgoing_beam.mu_s, outgoing_p_array.tau().mean(), atol=1e-7)
    assert np.isclose(outgoing_beam.mu_p, outgoing_p_array.p().mean())

    assert np.allclose(outgoing_beam.xs, outgoing_p_array.x())
    assert np.allclose(outgoing_beam.xps, outgoing_p_array.px())
    assert np.allclose(outgoing_beam.ys, outgoing_p_array.y())
    assert np.allclose(outgoing_beam.yps, outgoing_p_array.py())
    assert np.allclose(
        outgoing_beam.ss, outgoing_p_array.tau(), atol=1e-7, rtol=1e-1
    )  # TODO: Why do we need such large tolerances?
    assert np.allclose(outgoing_beam.ps, outgoing_p_array.p())


def test_twiss_particle_beam():
    """
    Test that the twiss parameters computed by Cheetah for a `ParticleBeam` loaded from
    an Astra beam are the same as those computed by Ocelot for the `ParticleArray`
    loaded from that same Astra beam.
    """
    # Cheetah
    particle_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )

    # Ocelot
    p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    ocelot_twiss = ocelot.cpbd.beam.get_envelope(p_array)

    # Compare
    assert np.isclose(particle_beam.emittance_x, ocelot_twiss.emit_x)
    assert np.isclose(particle_beam.normalized_emittance_x, ocelot_twiss.emit_xn)
    assert np.isclose(
        particle_beam.beta_x, ocelot_twiss.beta_x, rtol=1e-4
    )  # TODO: Is tolerance okay?
    assert np.isclose(particle_beam.alpha_x, ocelot_twiss.alpha_x, rtol=1e-4)
    assert np.isclose(particle_beam.emittance_y, ocelot_twiss.emit_y)
    assert np.isclose(particle_beam.normalized_emittance_y, ocelot_twiss.emit_yn)
    assert np.isclose(
        particle_beam.beta_y, ocelot_twiss.beta_y, rtol=1e-4
    )  # TODO: Is tolerance okay?
    assert np.isclose(particle_beam.alpha_y, ocelot_twiss.alpha_y, rtol=1e-4)


def test_twiss_parameter_beam():
    """
    Test that the twiss parameters computed by Cheetah for a `ParameterBeam` loaded from
    an Astra beam are the same as those computed by Ocelot for the `ParticleArray`
    loaded from that same Astra beam.
    """
    # Cheetah
    parameter_beam = cheetah.ParameterBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )

    # Ocelot
    p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    ocelot_twiss = ocelot.cpbd.beam.get_envelope(p_array)

    # Compare
    assert np.isclose(parameter_beam.emittance_x, ocelot_twiss.emit_x)
    assert np.isclose(parameter_beam.normalized_emittance_x, ocelot_twiss.emit_xn)
    assert np.isclose(
        parameter_beam.beta_x, ocelot_twiss.beta_x, rtol=1e-4
    )  # TODO: Is tolerance okay?
    assert np.isclose(parameter_beam.alpha_x, ocelot_twiss.alpha_x, rtol=1e-4)
    assert np.isclose(parameter_beam.emittance_y, ocelot_twiss.emit_y)
    assert np.isclose(parameter_beam.normalized_emittance_y, ocelot_twiss.emit_yn)
    assert np.isclose(
        parameter_beam.beta_y, ocelot_twiss.beta_y, rtol=1e-4
    )  # TODO: Is tolerance okay?
    assert np.isclose(parameter_beam.alpha_y, ocelot_twiss.alpha_y, rtol=1e-4)


def test_astra_import():
    """
    Test if the beam imported from Astra in Cheetah matches the beam imported from Astra
    in Ocelot.
    """
    beam = cheetah.ParticleBeam.from_astra("benchmark/astra/ACHIP_EA1_2021.1351.001")
    p_array = ocelot.astraBeam2particleArray("benchmark/astra/ACHIP_EA1_2021.1351.001")

    assert np.allclose(beam.particles[:, :6], p_array.rparticles.transpose())
    assert np.isclose(beam.energy, (p_array.E * 1e9))


def test_quadrupole():
    """
    Test if the tracking results through a Cheeath `Quadrupole` element match those
    through an Ocelot `Quadrupole` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_quadrupole = cheetah.Quadrupole(
        length=torch.tensor(0.23), k1=torch.tensor(5.0)
    )
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(0.1)),
            cheetah_quadrupole,
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_quadrupole = ocelot.Quadrupole(l=0.23, k1=5.0)
    lattice = ocelot.MagneticLattice(
        [ocelot.Drift(l=0.1), ocelot_quadrupole, ocelot.Drift(l=0.1)]
    )
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    # Split in order to allow for different tolerances for each particle dimension
    assert np.allclose(
        outgoing_beam.particles[:, 0], outgoing_p_array.rparticles.transpose()[:, 0]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 1], outgoing_p_array.rparticles.transpose()[:, 1]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 2], outgoing_p_array.rparticles.transpose()[:, 2]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 3], outgoing_p_array.rparticles.transpose()[:, 3]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 4],
        outgoing_p_array.rparticles.transpose()[:, 4],
        atol=1e-6,  # TODO: Is this tolerance already too large?
    )
    assert np.allclose(
        outgoing_beam.particles[:, 5], outgoing_p_array.rparticles.transpose()[:, 5]
    )


def test_tilted_quadrupole():
    """
    Test if the tracking results through a tilted Cheeath `Quadrupole` element match
    those through a tilted Ocelot `Quadrupole` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_quadrupole = cheetah.Quadrupole(
        length=torch.tensor(0.23), k1=torch.tensor(5.0), tilt=torch.tensor(0.79)
    )
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(0.1)),
            cheetah_quadrupole,
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    ocelot_quadrupole = ocelot.Quadrupole(l=0.23, k1=5.0, tilt=0.79)
    lattice = ocelot.MagneticLattice(
        [ocelot.Drift(l=0.1), ocelot_quadrupole, ocelot.Drift(l=0.1)]
    )
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(lattice, deepcopy(incoming_p_array), navigator)

    # Split in order to allow for different tolerances for each particle dimension
    assert np.allclose(
        outgoing_beam.particles[:, 0], outgoing_p_array.rparticles.transpose()[:, 0]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 1], outgoing_p_array.rparticles.transpose()[:, 1]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 2], outgoing_p_array.rparticles.transpose()[:, 2]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 3], outgoing_p_array.rparticles.transpose()[:, 3]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 4],
        outgoing_p_array.rparticles.transpose()[:, 4],
        atol=1e-6,  # TODO: Is this tolerance already too large?
    )
    assert np.allclose(
        outgoing_beam.particles[:, 5], outgoing_p_array.rparticles.transpose()[:, 5]
    )


def test_sbend():
    """
    Test if the tracking results through a Cheeath `Dipole` element match those through
    an Ocelot `SBend` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_dipole = cheetah.Dipole(length=torch.tensor(0.1), angle=torch.tensor(2e-5))
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(0.1)),
            cheetah_dipole,
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    ocelot_sbend = ocelot.SBend(l=0.1, angle=2e-5)
    lattice = ocelot.MagneticLattice(
        [ocelot.Drift(l=0.1), ocelot_sbend, ocelot.Drift(l=0.1)]
    )
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(
        lattice, deepcopy(incoming_p_array), navigator, print_progress=False
    )

    # Split in order to allow for different tolerances for each particle dimension
    assert np.allclose(
        outgoing_beam.particles[:, 0],
        outgoing_p_array.rparticles.transpose()[:, 0],
        rtol=1e-3,
    )
    assert np.allclose(
        outgoing_beam.particles[:, 1], outgoing_p_array.rparticles.transpose()[:, 1]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 2], outgoing_p_array.rparticles.transpose()[:, 2]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 3], outgoing_p_array.rparticles.transpose()[:, 3]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 4],
        outgoing_p_array.rparticles.transpose()[:, 4],
        atol=1e-6,  # TODO: Is this tolerance already too large?
    )
    assert np.allclose(
        outgoing_beam.particles[:, 5], outgoing_p_array.rparticles.transpose()[:, 5]
    )


def test_rbend():
    """
    Test if the tracking results through a Cheeath `RBend` element match those through
    an Ocelot `RBend` element.
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_dipole = cheetah.RBend(length=torch.tensor(0.1), angle=torch.tensor(2e-5))
    cheetah_segment = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(0.1)),
            cheetah_dipole,
            cheetah.Drift(length=torch.tensor(0.1)),
        ]
    )
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    ocelot_rbend = ocelot.RBend(l=0.1, angle=2e-5)
    lattice = ocelot.MagneticLattice(
        [ocelot.Drift(l=0.1), ocelot_rbend, ocelot.Drift(l=0.1)]
    )
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(
        lattice, deepcopy(incoming_p_array), navigator, print_progress=False
    )

    # Split in order to allow for different tolerances for each particle dimension
    assert np.allclose(
        outgoing_beam.particles[:, 0],
        outgoing_p_array.rparticles.transpose()[:, 0],
        rtol=1e-3,
    )
    assert np.allclose(
        outgoing_beam.particles[:, 1], outgoing_p_array.rparticles.transpose()[:, 1]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 2], outgoing_p_array.rparticles.transpose()[:, 2]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 3], outgoing_p_array.rparticles.transpose()[:, 3]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 4],
        outgoing_p_array.rparticles.transpose()[:, 4],
        atol=1e-6,  # TODO: Is this tolerance already too large?
    )
    assert np.allclose(
        outgoing_beam.particles[:, 5], outgoing_p_array.rparticles.transpose()[:, 5]
    )


def test_convert_rbend():
    """
    Test if the tracking results through a ocelot-converted Cheetah segment match
    those through an Ocelot section with an `RBend` element.
    """
    # Ocelot
    incoming_p_array = ocelot.astraBeam2particleArray(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001", print_params=False
    )
    ocelot_rbend = ocelot.RBend(l=0.1, angle=2e-5, gap=0.05, fint=0.1, eid="rbend")
    lattice = ocelot.MagneticLattice(
        [
            ocelot.Drift(l=0.1, eid="drfit_1"),
            ocelot_rbend,
            ocelot.Drift(l=0.1, eid="drift_2"),
        ]
    )
    navigator = ocelot.Navigator(lattice)
    _, outgoing_p_array = ocelot.track(
        lattice, deepcopy(incoming_p_array), navigator, print_progress=False
    )

    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/cheetah/ACHIP_EA1_2021.1351.001"
    )
    cheetah_segment = cheetah.Segment.from_ocelot(lattice.sequence)
    outgoing_beam = cheetah_segment.track(incoming_beam)

    # Split in order to allow for different tolerances for each particle dimension
    assert np.allclose(
        outgoing_beam.particles[:, 0],
        outgoing_p_array.rparticles.transpose()[:, 0],
        rtol=1e-3,
    )
    assert np.allclose(
        outgoing_beam.particles[:, 1], outgoing_p_array.rparticles.transpose()[:, 1]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 2], outgoing_p_array.rparticles.transpose()[:, 2]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 3], outgoing_p_array.rparticles.transpose()[:, 3]
    )
    assert np.allclose(
        outgoing_beam.particles[:, 4],
        outgoing_p_array.rparticles.transpose()[:, 4],
        atol=1e-6,  # TODO: Is this tolerance already too large?
    )
    assert np.allclose(
        outgoing_beam.particles[:, 5], outgoing_p_array.rparticles.transpose()[:, 5]
    )


def test_cavity():
    """
    Compare tracking through a cavity that is on.

    The particular settings tested here also get to the same result in Bmad/Tao. The
    below lattice was used for the Bmad test:
    ```
    ! Lattice file: simple.bmad
    beginning[beta_a] = 5.91253677 ! m a-mode beta function
    beginning[beta_b] = 5.91253677 ! m b-mode beta function
    beginning[alpha_a] = 3.55631308 ! a-mode alpha function
    beginning[alpha_b] = 3.55631308 ! b-mode alpha function
    beginning[e_tot] = 6e6 ! eV    Or can set beginning[p0c]

    parameter[geometry] = open  ! Or closed
    parameter[particle] = electron  ! Reference particle.

    c: lcavity, rf_frequency = 1.3e9, l = 1.0377, voltage = 0.01815975e9, phi0 = 0.0

    lat: line = (c) ! List of lattice elements
    use, lat ! Line used to construct the lattice
    ```
    The twiss parameters at the end of the lattice according to Bmad should be:
     - beta_x  = 0.23847352510683092
     - beta_y  = 0.23847352512430994
     - alpha_x = -1.0160687592932345
     - alpha_y = -1.0160687593664295
    """
    # Cheetah
    incoming_beam = cheetah.ParticleBeam.from_twiss(
        beta_x=torch.tensor(5.91253677),
        alpha_x=torch.tensor(3.55631308),
        beta_y=torch.tensor(5.91253677),
        alpha_y=torch.tensor(3.55631308),
        emittance_x=torch.tensor(3.494768647122823e-09),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )
    cheetah_cavity = cheetah.Cavity(
        length=torch.tensor(1.0377),
        voltage=torch.tensor(0.01815975e9),
        frequency=torch.tensor(1.3e9),
        phase=torch.tensor(0.0),
    )
    outgoing_beam = cheetah_cavity.track(incoming_beam)

    # Ocelot
    tws = ocelot.Twiss()
    tws.beta_x = 5.91253677
    tws.alpha_x = 3.55631308
    tws.beta_y = 5.91253677
    tws.alpha_y = 3.55631308
    tws.emit_x = 3.494768647122823e-09
    tws.emit_y = 3.497810737006068e-09
    tws.gamma_x = (1 + tws.alpha_x**2) / tws.beta_x
    tws.gamma_y = (1 + tws.alpha_y**2) / tws.beta_y
    tws.E = 6e-3

    p_array = ocelot.generate_parray(tws=tws)

    cell = [ocelot.Cavity(l=1.0377, v=0.01815975, freq=1.3e9, phi=0.0)]
    lattice = ocelot.MagneticLattice(cell)
    navigator = ocelot.Navigator(lattice=lattice)

    _, outgoing_parray = ocelot.track(lattice, deepcopy(p_array), navigator)
    derived_twiss = ocelot.cpbd.beam.get_envelope(outgoing_parray)

    # Compare
    assert np.isclose(outgoing_beam.beta_x, derived_twiss.beta_x, rtol=1e-2)
    assert np.isclose(outgoing_beam.alpha_x, derived_twiss.alpha_x, rtol=1e-2)
    assert np.isclose(outgoing_beam.beta_y, derived_twiss.beta_y, rtol=1e-2)
    assert np.isclose(outgoing_beam.alpha_y, derived_twiss.alpha_y, rtol=1e-2)



from cheetah.accelerator import Segment

from .resources import ARESlatticeStage3v1_9 as ares


def test_save_and_reload(tmp_path):
    """
    Test that saving Cheetah `Segment` to LatticeJSON works and that it can be reloaded
    correctly.
    """
    original_segment = Segment.from_ocelot(ares.cell, name="ARES_Segment")

    original_segment.to_lattice_json(
        str(tmp_path / "ares_lattice.json"),
        title="ARES LatticeJSON",
        info="Save and reload test for Cheetah using the ARES lattice",
    )

    reloaded_segment = Segment.from_lattice_json(str(tmp_path / "ares_lattice.json"))

    assert original_segment.name == reloaded_segment.name
    assert len(original_segment.elements) == len(reloaded_segment.elements)
    assert original_segment.length == reloaded_segment.length

    # TODO: Improve element comparison when equality is working again
    for original_element, reloaded_element in zip(
        original_segment.elements, reloaded_segment.elements
    ):
        assert original_element.name == reloaded_element.name
        assert original_element.__class__ == reloaded_element.__class__



import time

import torch

import cheetah

from .resources import ARESlatticeStage3v1_9 as ares


# TODO: Test that Cheeath tracks faster than Ocelot
def test_tracking_speed():
    """Really only tests that Cheetah isn't super slow."""
    cell = cheetah.converters.nocelot.subcell_of_ocelot(
        ares.cell, "AREASOLA1", "AREABSCR1"
    )
    segment = cheetah.Segment.from_ocelot(cell)
    segment.AREABSCR1.is_active = True  # Turn screen on and off

    particles = cheetah.ParticleBeam.from_parameters(
        num_particles=torch.tensor(int(1e5)),
        sigma_x=torch.tensor(175e-6),
        sigma_y=torch.tensor(175e-6),
    )

    t1 = time.time()

    _ = segment.track(particles)
    _ = segment.AREABSCR1.reading

    t2 = time.time()

    assert (t2 - t1) < 0.1



import json
from typing import Any, Optional, Tuple

import torch

import cheetah


def feature2nontorch(value: Any) -> Any:
    """
    if necesary, convert an the value of a feature of a `cheetah.Element` to a non-torch
    type that can be saved to LatticeJSON.

    :param value: Value of the feature that might be in some kind of PyTorch format,
        such as `torch.Tensor` or `torch.nn.Parameter`.
    :return: Value of the feature if it is not in a PyTorch format, otherwise the
        value converted to a non-PyTorch format.
    """
    return (
        value.tolist()
        if isinstance(value, (torch.Tensor, torch.nn.Parameter))
        else value
    )


def convert_element(element: "cheetah.Element"):
    """
    Deconstruct an element into its name, class and parameters for saving to JSON.

    :param element: Cheetah element
    :return: Tuple of element name, element class, and element parameters
    """
    params = {
        feauture: feature2nontorch(getattr(element, feauture))
        for feauture in element.defining_features
    }

    return element.name, element.__class__.__name__, params


def convert_segment(segment: "cheetah.Segment") -> Tuple[dict, dict]:
    """
    Deconstruct a segment into its name, a list of its elements and a dictionary of
    its element parameters for saving to JSON.

    :param segment: Cheetah segment.
    :return: Tuple of elments and lattices dictionaries found in segment, including
        the segment itself.
    """
    elements = {}
    lattices = {}

    cell = []

    for element in segment.elements:
        if isinstance(element, cheetah.Segment):
            segment_elements, segment_lattices = convert_segment(element)

            elements.update(segment_elements)
            lattices.update(segment_lattices)
        else:
            element_name, element_class, element_params = convert_element(element)

            elements[element_name] = [element_class, element_params]

        cell.append(element_name)

    lattices[segment.name] = cell

    return elements, lattices


def save_cheetah_model(
    segment: "cheetah.Segment",
    filename: str,
    title: Optional[str] = None,
    info: str = "This is a placeholder lattice description",
) -> None:
    """
    Save a cheetah model to json file accoding to the lattice-json convention
    c.f. https://github.com/nobeam/latticejson

    :param segment: Cheetah `Segment` to save.
    :param filename: Name/path of the file to save the lattice to.
    :param title: Title of the lattice. If not provided, defaults to the name of the
        `Segment` object. If that also does not have a name, defaults to "Unnamed
        Lattice".
    :param info: Information about the lattice. Defaults to "This is a placeholder
        lattice description".
    """
    if title is None:
        title = segment.name if segment.name is not None else "Unnamed Lattice"

    metadata = {
        "version": "cheetah-0.6",
        "title": title,
        "info": info,
        "root": segment.name if segment.name is not None else "cell",
    }

    lattice_dict = metadata.copy()

    elements, lattices = convert_segment(segment)
    lattice_dict["elements"] = elements
    lattice_dict["lattices"] = lattices

    with open(filename, "w") as f:
        s = json.dumps(lattice_dict, cls=CompactJSONEncoder, indent=4)
        f.write(s)


class CompactJSONEncoder(json.JSONEncoder):
    """
    A JSON Encoder which only indents the first two levels.

    Taken from https://github.com/nobeam/latticejson/blob/main/latticejson/format.py
    """

    def encode(self, obj, level=0):
        if isinstance(obj, dict) and level < 2:
            items_indent = (level + 1) * self.indent * " "
            items_string = ",\n".join(
                f"{items_indent}{json.dumps(key)}: {self.encode(value, level=level+1)}"
                for key, value in obj.items()
            )
            dict_indent = level * self.indent * " "
            newline = "\n" if level == 0 else ""
            return f"{{\n{items_string}\n{dict_indent}}}{newline}"
        else:
            return json.dumps(obj)


def nontorch2feature(value: Any) -> Any:
    """
    Convert a value like a float, int, etc. to a torch tensor if necessary. Values of
    type `str` and `bool` are not converted, because the all currently existing
    `cheetah.Element` subclasses expect these values to not be `torch.Tensor`s.

    :param value: Value to convert to a `torch.Tensor` if necessary.
    :return: Value converted to a `torch.Tensor` if necessary.
    """
    return value if isinstance(value, (str, bool)) else torch.tensor(value)


def parse_element(name: str, lattice_dict: dict) -> "cheetah.Element":
    """
    Parse an `Element` named `name` from a `lattice_dict`.

    :param name: Name of the `Element` to parse.
    :param lattice_dict: Dictionary containing the lattice information.
    """
    element_class = getattr(cheetah, lattice_dict["elements"][name][0])
    params = lattice_dict["elements"][name][1]

    converted_params = {key: nontorch2feature(value) for key, value in params.items()}

    return element_class(name=name, **converted_params)


def parse_segment(name: str, lattice_dict: dict) -> "cheetah.Segment":
    """
    Parse a `Segment` named `name` from a `lattice_dict`.

    :param name: Name of the `Segment` to parse.
    :param lattice_dict: Dictionary containing the lattice information.
    """
    elements = []
    for element_name in lattice_dict["lattices"][name]:
        # Construct new element
        if element_name in lattice_dict["lattices"]:
            new_element = parse_segment(element_name, lattice_dict)
        else:
            new_element = parse_element(element_name, lattice_dict)

        # Append the element to the list of elements
        elements.append(new_element)

    return cheetah.Segment(elements=elements, name=name)


def load_cheetah_model(filename: str) -> "cheetah.Segment":
    """
    Load a Cheetah model from a JSON file.

    :param filename: Name/path of the file to load the lattice from.
    :return: Loaded Cheetah `Segment`.
    """
    with open(filename, "r") as f:
        lattice_dict = json.load(f)

    root_name = lattice_dict["root"]

    return parse_segment(root_name, lattice_dict)



from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Any, Literal, Optional, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.patches import Rectangle
from scipy import constants
from scipy.constants import physical_constants
from scipy.stats import multivariate_normal
from torch import nn

from cheetah.converters.dontbmad import convert_bmad_lattice
from cheetah.error import DeviceError
from cheetah.latticejson import load_cheetah_model, save_cheetah_model
from cheetah.particles import Beam, ParameterBeam, ParticleBeam
from cheetah.track_methods import base_rmatrix, misalignment_matrix, rotation_matrix
from cheetah.utils import UniqueNameGenerator

generate_unique_name = UniqueNameGenerator(prefix="unnamed_element")

rest_energy = torch.tensor(
    constants.electron_mass
    * constants.speed_of_light**2
    / constants.elementary_charge
)  # electron mass
electron_mass_eV = torch.tensor(
    physical_constants["electron mass energy equivalent in MeV"][0] * 1e6
)


class Element(ABC, nn.Module):
    """
    Base class for elements of particle accelerators.

    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(self, name: Optional[str] = None, device: str = "auto") -> None:
        super().__init__()

        self.name = name if name is not None else generate_unique_name()

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        """
        Generates the element's transfer map that describes how the beam and its
        particles are transformed when traveling through the element. The state vector
        consists of 6 values with a physical meaning:
        - x: Position in x direction
        - xp: Momentum in x direction
        - y: Position in y direction
        - yp: Momentum in y direction
        - s: Position in z direction, the zero value is set to the middle of the pulse
        - sp: Momentum in s direction
        As well as a seventh value used to add constants to some of the prior values if
        necessary. Through this seventh state, the addition of constants can be
        represented using a matrix multiplication.

        :param energy: Energy of the Beam. Read from the fed-in Cheetah Beam.
        :return: A 7x7 Matrix for further calculations.
        """
        raise NotImplementedError

    def track(self, incoming: Beam) -> Beam:
        """
        Track particles through the element. The input can be a `ParameterBeam` or a
        `ParticleBeam`.

        :param incoming: Beam of particles entering the element.
        :return: Beam of particles exiting the element.
        """
        if incoming is Beam.empty:
            return incoming
        elif isinstance(incoming, ParameterBeam):
            if self.device != incoming.device:
                raise DeviceError
            tm = self.transfer_map(incoming.energy)
            mu = torch.matmul(tm, incoming._mu)
            cov = torch.matmul(tm, torch.matmul(incoming._cov, tm.t()))
            return ParameterBeam(mu, cov, incoming.energy, device=incoming.device)
        elif isinstance(incoming, ParticleBeam):
            if self.device != incoming.device:
                raise DeviceError
            tm = self.transfer_map(incoming.energy)
            new_particles = torch.matmul(incoming.particles, tm.t())
            return ParticleBeam(new_particles, incoming.energy, device=incoming.device)
        else:
            raise TypeError(f"Parameter incoming is of invalid type {type(incoming)}")

    def forward(self, incoming: Beam) -> Beam:
        """Forward function required by `torch.nn.Module`. Simply calls `track`."""
        return self.track(incoming)

    @property
    @abstractmethod
    def is_skippable(self) -> bool:
        """
        Whether the element can be skipped during tracking. If `True`, the element's
        transfer map is combined with the transfer maps of surrounding skipable
        elements.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def defining_features(self) -> list[str]:
        """
        List of features that define the element. Used to compare elements for equality
        and to save them.

        NOTE: When overriding this property, make sure to call the super method and
        extend the list it returns.
        """
        return []

    @abstractmethod
    def split(self, resolution: torch.Tensor) -> list["Element"]:
        """
        Split the element into slices no longer than `resolution`. Some elements may not
        be splittable, in which case a list containing only the element itself is
        returned.

        :param resolution: Length of the longest allowed split in meters.
        :return: Ordered sequence of sliced elements.
        """
        raise NotImplementedError

    @abstractmethod
    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        """
        Plot a representation of this element into a `matplotlib` Axes at position `s`.

        :param ax: Axes to plot the representation into.
        :param s: Position of the object along s in meters.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={repr(self.name)},"
            f" device={repr(self.device)})"
        )


class Drift(Element):
    """
    Drift section in a particle accelerator.

    :param length: Length in meters.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        gamma = energy / rest_energy
        igamma2 = 1 / gamma**2 if gamma != 0 else torch.tensor(0.0)

        tm = torch.eye(7, device=self.device)
        tm[0, 1] = self.length
        tm[2, 3] = self.length
        tm[4, 5] = self.length * igamma2

        return tm

    @property
    def is_skippable(self) -> bool:
        return True

    def split(self, resolution: torch.Tensor) -> list[Element]:
        split_elements = []
        remaining = self.length
        while remaining > 0:
            element = Drift(torch.min(resolution, remaining), device=self.device)
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        pass

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)},"
            f" name={repr(self.name)}, device={repr(self.device)})"
        )


class Quadrupole(Element):
    """
    Quadrupole magnet in a particle accelerator.

    :param length: Length in meters.
    :param k1: Strength of the quadrupole in rad/m.
    :param misalignment: Misalignment vector of the quadrupole in x- and y-directions.
    :param tilt: Tilt angle of the quadrupole in x-y plane [rad]. pi/4 for
        skew-quadrupole.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        k1: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        misalignment: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        tilt: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length
        self.k1 = k1 if k1 is not None else torch.tensor(0.0)
        self.misalignment = (
            misalignment if misalignment is not None else torch.tensor([0.0, 0.0])
        )
        self.tilt = tilt if tilt is not None else torch.tensor(0.0)

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        R = base_rmatrix(
            length=self.length,
            k1=self.k1,
            hx=torch.tensor(0.0),
            tilt=self.tilt,
            energy=energy,
            device=self.device,
        )

        if self.misalignment[0] == 0 and self.misalignment[1] == 0:
            return R
        else:
            R_exit, R_entry = misalignment_matrix(self.misalignment, self.device)
            R = torch.matmul(R_exit, torch.matmul(R, R_entry))
            return R

    @property
    def is_skippable(self) -> bool:
        return True

    @property
    def is_active(self) -> bool:
        return self.k1 != 0

    def split(self, resolution: torch.Tensor) -> list[Element]:
        split_elements = []
        remaining = self.length
        while remaining > 0:
            element = Quadrupole(
                torch.min(resolution, remaining),
                self.k1,
                misalignment=self.misalignment,
                device=self.device,
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.k1) if self.is_active else 1)
        patch = Rectangle(
            (s, 0), self.length, height, color="tab:red", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length", "k1", "misalignment", "tilt"]

    def __repr__(self) -> None:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"k1={repr(self.k1)}, "
            + f"misalignment={repr(self.misalignment)}, "
            + f"tilt={repr(self.tilt)}, "
            + f"name={repr(self.name)}, "
            + f'device="{repr(self.device)}")'
        )


class Dipole(Element):
    """
    Dipole magnet (by default a sector bending magnet).

    :param length: Length in meters.
    :param angle: Deflection angle in rad.
    :param e1: The angle of inclination of the entrance face [rad].
    :param e2: The angle of inclination of the exit face [rad].
    :param tilt: Tilt of the magnet in x-y plane [rad].
    :param fringe_integral: Fringe field integral (of the enterance face).
    :param fringe_integral_exit: (only set if different from `fint`) Fringe field
        integral of the exit face.
    :param gap: The magnet gap [m], NOTE in MAD and ELEGANT: HGAP = gap/2
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        angle: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        e1: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        e2: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        tilt: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        fringe_integral: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        fringe_integral_exit: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        gap: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ):
        super().__init__(name=name, device=device)

        self.length = length
        self.angle = angle if angle is not None else torch.tensor(0.0)
        self.gap = gap if gap is not None else torch.tensor(0.0)
        self.tilt = tilt if tilt is not None else torch.tensor(0.0)
        self.name = name
        self.fringe_integral = (
            fringe_integral if fringe_integral is not None else torch.tensor(0.0)
        )
        self.fringe_integral_exit = (
            fringe_integral if fringe_integral_exit is None else fringe_integral_exit
        )
        # Rectangular bend
        self.e1 = e1 if e1 is not None else torch.tensor(0.0)
        self.e2 = e2 if e2 is not None else torch.tensor(0.0)

        if self.length == 0.0:
            self.hx = torch.tensor(0.0)
        else:
            self.hx = self.angle / self.length

    @property
    def is_skippable(self) -> bool:
        return True

    @property
    def is_active(self):
        return self.angle != 0

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        R_enter = self._transfer_map_enter(energy)
        R_exit = self._transfer_map_exit(energy)

        if self.length != 0.0:  # Bending magnet with finite length
            R = base_rmatrix(
                length=self.length,
                k1=torch.tensor(0.0),
                hx=self.hx,
                tilt=torch.tensor(0.0),
                energy=energy,
                device=self.device,
            )
        else:  # Reduce to Thin-Corrector
            R = torch.eye(7, device=self.device)
            R[0, 1] = self.length
            R[2, 6] = self.angle
            R[2, 3] = self.length

        # Apply fringe fields
        R = torch.matmul(R_exit, torch.matmul(R, R_enter))
        # Apply rotation for tilted magnets
        R = torch.matmul(
            rotation_matrix(-self.tilt), torch.matmul(R, rotation_matrix(self.tilt))
        )

        return R

    def _transfer_map_enter(self, energy: torch.Tensor) -> torch.Tensor:
        if self.fringe_integral == 0:
            return torch.eye(7, device=self.device)
        else:
            sec_e = torch.tensor(1.0) / torch.cos(self.e1)
            phi = (
                self.fringe_integral
                * self.hx
                * self.gap
                * sec_e
                * (1 + torch.sin(self.e1) ** 2)
            )

            tm = torch.eye(7, device=self.device)
            tm[1, 0] = self.hx * torch.tan(self.e1)
            tm[3, 2] = -self.hx * torch.tan(self.e1 - phi)

            return tm

    def _transfer_map_exit(self, energy: torch.Tensor) -> torch.Tensor:
        if self.fringe_integral_exit == 0:
            return torch.eye(7, device=self.device)
        else:
            sec_e = 1.0 / torch.cos(self.e2)
            phi = (
                self.fringe_integral
                * self.hx
                * self.gap
                * sec_e
                * (1 + torch.sin(self.e2) ** 2)
            )

            tm = torch.eye(7, device=self.device)
            tm[1, 0] = self.hx * torch.tan(self.e2)
            tm[3, 2] = -self.hx * torch.tan(self.e2 - phi)

            return tm

    def split(self, resolution: torch.Tensor) -> list[Element]:
        # TODO: Implement splitting for dipole properly, for now just returns the
        # element itself
        return [self]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"angle={repr(self.angle)}, "
            + f"e1={repr(self.e1)},"
            + f"e2={repr(self.e2)},"
            + f"tilt={repr(self.tilt)},"
            + f"fringe_integral={repr(self.fringe_integral)},"
            + f"fringe_integral_exit={repr(self.fringe_integral_exit)},"
            + f"gap={repr(self.gap)},"
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + [
            "length",
            "angle",
            "e1",
            "e2",
            "tilt",
            "fringe_integral",
            "fringe_integral_exit",
            "gap",
        ]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.angle) if self.is_active else 1)

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:green", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)


class RBend(Dipole):
    """
    Rectangular bending magnet.

    :param length: Length in meters.
    :param angle: Deflection angle in rad.
    :param e1: The angle of inclination of the entrance face [rad].
    :param e2: The angle of inclination of the exit face [rad].
    :param tilt: Tilt of the magnet in x-y plane [rad].
    :param fringe_integral: Fringe field integral (of the enterance face).
    :param fringe_integral_exit: (only set if different from `fint`) Fringe field
        integral of the exit face.
    :param gap: The magnet gap [m], NOTE in MAD and ELEGANT: HGAP = gap/2
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Optional[Union[torch.Tensor, nn.Parameter]],
        angle: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        e1: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        e2: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        tilt: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        fringe_integral: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        fringe_integral_exit: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        gap: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ):
        angle = angle if angle is not None else torch.tensor(0.0)
        e1 = e1 if e1 is not None else torch.tensor(0.0)
        e2 = e2 if e2 is not None else torch.tensor(0.0)
        tilt = tilt if tilt is not None else torch.tensor(0.0)
        fringe_integral = (
            fringe_integral if fringe_integral is not None else torch.tensor(0.0)
        )
        # fringe_integral_exit is left out on purpose
        gap = gap if gap is not None else torch.tensor(0.0)

        e1 = e1 + angle / 2
        e2 = e2 + angle / 2

        super().__init__(
            length=length,
            angle=angle,
            e1=e1,
            e2=e2,
            tilt=tilt,
            fringe_integral=fringe_integral,
            fringe_integral_exit=fringe_integral_exit,
            gap=gap,
            name=name,
            device=device,
        )


class HorizontalCorrector(Element):
    """
    Horizontal corrector magnet in a particle accelerator.

    :param length: Length in meters.
    :param angle: Particle deflection angle in the horizontal plane in rad.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        angle: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length
        self.angle = angle if angle is not None else torch.tensor(0.0)

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        tm = torch.eye(7, device=self.device)
        tm[0, 1] = self.length
        tm[1, 6] = self.angle
        tm[2, 3] = self.length

        return tm

    @property
    def is_skippable(self) -> bool:
        return True

    @property
    def is_active(self) -> bool:
        return self.angle != 0

    def split(self, resolution: torch.Tensor) -> list[Element]:
        split_elements = []
        remaining = self.length
        while remaining > 0:
            length = torch.min(resolution, remaining)
            element = HorizontalCorrector(
                length, self.angle * length / self.length, device=self.device
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.angle) if self.is_active else 1)

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:blue", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length", "angle"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"angle={repr(self.angle)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class VerticalCorrector(Element):
    """
    Verticle corrector magnet in a particle accelerator.

    :param length: Length in meters.
    :param angle: Particle deflection angle in the vertical plane in rad.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        angle: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length
        self.angle = angle if angle is not None else torch.tensor(0.0)

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        tm = torch.eye(7, device=self.device)
        tm[0, 1] = self.length
        tm[2, 3] = self.length
        tm[3, 6] = self.angle

        return tm

    @property
    def is_skippable(self) -> bool:
        return True

    @property
    def is_active(self) -> bool:
        return self.angle != 0

    def split(self, resolution: torch.Tensor) -> list[Element]:
        split_elements = []
        remaining = self.length
        while remaining > 0:
            length = torch.min(resolution, remaining)
            element = VerticalCorrector(
                length, self.angle * length / self.length, device=self.device
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.angle) if self.is_active else 1)

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:cyan", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length", "angle"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"angle={repr(self.angle)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class Cavity(Element):
    """
    Accelerating cavity in a particle accelerator.

    :param length: Length in meters.
    :param voltage: Voltage of the cavity in volts.
    :param phase: Phase of the cavity in degrees.
    :param frequency: Frequency of the cavity in Hz.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        voltage: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        phase: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        frequency: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length
        self.voltage = voltage if voltage is not None else torch.tensor(0.0)
        self.phase = phase if phase is not None else torch.tensor(0.0)
        self.frequency = frequency if frequency is not None else torch.tensor(0.0)

    @property
    def is_active(self) -> bool:
        return self.voltage != 0

    @property
    def is_skippable(self) -> bool:
        return not self.is_active

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        if self.voltage > 0:
            return self._cavity_rmatrix(energy)
        else:
            return base_rmatrix(
                length=self.length,
                k1=torch.tensor(0.0),
                hx=torch.tensor(0.0),
                tilt=torch.tensor(0.0),
                energy=energy,
                device=self.device,
            )

    def track(self, incoming: Beam) -> Beam:
        """
        Track particles through the cavity. The input can be a `ParameterBeam` or a
        `ParticleBeam`. For a cavity, this does a little more than just the transfer map
        multiplication done by most elements.

        :param incoming: Beam of particles entering the element.
        :return: Beam of particles exiting the element.
        """
        if incoming is Beam.empty:
            return incoming
        elif isinstance(incoming, (ParameterBeam, ParticleBeam)):
            return self._track_beam(incoming)
        else:
            raise TypeError(f"Parameter incoming is of invalid type {type(incoming)}")

    def _track_beam(self, incoming: ParticleBeam) -> ParticleBeam:
        beta0 = torch.tensor(1.0)
        igamma2 = torch.tensor(0.0)
        g0 = torch.tensor(1e10)
        if incoming.energy != 0:
            g0 = incoming.energy / electron_mass_eV
            igamma2 = 1 / g0**2
            beta0 = torch.sqrt(1 - igamma2)

        phi = torch.deg2rad(self.phase)

        tm = self.transfer_map(incoming.energy)
        if isinstance(incoming, ParameterBeam):
            outgoing_mu = torch.matmul(tm, incoming._mu)
            outgoing_cov = torch.matmul(tm, torch.matmul(incoming._cov, tm.t()))
        else:  # ParticleBeam
            outgoing_particles = torch.matmul(incoming.particles, tm.t())
        delta_energy = self.voltage * torch.cos(phi)

        T566 = 1.5 * self.length * igamma2 / beta0**3
        T556 = 0
        T555 = 0
        if incoming.energy + delta_energy > 0:
            k = 2 * torch.pi * self.frequency / constants.speed_of_light
            outgoing_energy = incoming.energy + delta_energy
            g1 = outgoing_energy / electron_mass_eV
            beta1 = torch.sqrt(1 - 1 / g1**2)

            if isinstance(incoming, ParameterBeam):
                outgoing_mu[5] = (
                    incoming._mu[5]
                    + incoming.energy * beta0 / (outgoing_energy * beta1)
                    + self.voltage
                    * beta0
                    / (outgoing_energy * beta1)
                    * (torch.cos(incoming._mu[4] * beta0 * k + phi) - torch.cos(phi))
                )
                outgoing_cov[5, 5] = incoming._cov[5, 5]
                # outgoing_cov[5, 5] = (
                #     incoming._cov[5, 5]
                #     + incoming.energy * beta0 / (outgoing_energy * beta1)
                #     + self.voltage
                #     * beta0
                #     / (outgoing_energy * beta1)
                #     * (torch.cos(incoming._mu[4] * beta0 * k + phi) - torch.cos(phi))
                # )
            else:  # ParticleBeam
                outgoing_particles[:, 5] = (
                    incoming.particles[:, 5]
                    + incoming.energy * beta0 / (outgoing_energy * beta1)
                    + self.voltage
                    * beta0
                    / (outgoing_energy * beta1)
                    * (
                        torch.cos(incoming.particles[:, 4] * beta0 * k + phi)
                        - torch.cos(phi)
                    )
                )

            dgamma = self.voltage / electron_mass_eV
            if delta_energy > 0:
                T566 = (
                    self.length
                    * (beta0**3 * g0**3 - beta1**3 * g1**3)
                    / (2 * beta0 * beta1**3 * g0 * (g0 - g1) * g1**3)
                )
                T556 = (
                    beta0
                    * k
                    * self.length
                    * dgamma
                    * g0
                    * (beta1**3 * g1**3 + beta0 * (g0 - g1**3))
                    * torch.sin(phi)
                    / (beta1**3 * g1**3 * (g0 - g1) ** 2)
                )
                T555 = (
                    beta0**2
                    * k**2
                    * self.length
                    * dgamma
                    / 2.0
                    * (
                        dgamma
                        * (
                            2 * g0 * g1**3 * (beta0 * beta1**3 - 1)
                            + g0**2
                            + 3 * g1**2
                            - 2
                        )
                        / (beta1**3 * g1**3 * (g0 - g1) ** 3)
                        * torch.sin(phi) ** 2
                        - (g1 * g0 * (beta1 * beta0 - 1) + 1)
                        / (beta1 * g1 * (g0 - g1) ** 2)
                        * torch.cos(phi)
                    )
                )

            if isinstance(incoming, ParameterBeam):
                outgoing_mu[4] = (
                    T566 * incoming._mu[5] ** 2
                    + T556 * incoming._mu[4] * incoming._mu[5]
                    + T555 * incoming._mu[4] ** 2
                )
                outgoing_cov[4, 4] = (
                    T566 * incoming._cov[5, 5] ** 2
                    + T556 * incoming._cov[4, 5] * incoming._cov[5, 5]
                    + T555 * incoming._cov[4, 4] ** 2
                )
                outgoing_cov[4, 5] = (
                    T566 * incoming._cov[5, 5] ** 2
                    + T556 * incoming._cov[4, 5] * incoming._cov[5, 5]
                    + T555 * incoming._cov[4, 4] ** 2
                )
                outgoing_cov[5, 4] = outgoing_cov[4, 5]
            else:  # ParticleBeam
                outgoing_particles[:, 4] = (
                    T566 * incoming.particles[:, 5] ** 2
                    + T556 * incoming.particles[:, 4] * incoming.particles[:, 5]
                    + T555 * incoming.particles[:, 4] ** 2
                )

        if isinstance(incoming, ParameterBeam):
            outgoing = ParameterBeam(outgoing_mu, outgoing_cov, outgoing_energy)
            return outgoing
        else:  # ParticleBeam
            outgoing = ParticleBeam(
                outgoing_particles, outgoing_energy, device=incoming.device
            )
            return outgoing

    def _cavity_rmatrix(self, energy: torch.Tensor) -> torch.Tensor:
        """Produces an R-matrix for a cavity when it is on, i.e. voltage > 0.0."""
        phi = torch.deg2rad(self.phase)
        delta_energy = self.voltage * torch.cos(phi)
        # Comment from Ocelot: Pure pi-standing-wave case
        eta = torch.tensor(1.0)
        Ei = energy / electron_mass_eV
        Ef = (energy + delta_energy) / electron_mass_eV
        Ep = (Ef - Ei) / self.length  # Derivative of the energy
        assert Ei > 0, "Initial energy must be larger than 0"

        alpha = torch.sqrt(eta / 8) / torch.cos(phi) * torch.log(Ef / Ei)

        r11 = torch.cos(alpha) - torch.sqrt(2 / eta) * torch.cos(phi) * torch.sin(alpha)

        # In Ocelot r12 is defined as below only if abs(Ep) > 10, and self.length
        # otherwise. This is implemented differently here in order to achieve results
        # closer to Bmad.
        r12 = torch.sqrt(8 / eta) * Ei / Ep * torch.cos(phi) * torch.sin(alpha)

        r21 = (
            -Ep
            / Ef
            * (
                torch.cos(phi) / torch.sqrt(2 * eta)
                + torch.sqrt(eta / 8) / torch.cos(phi)
            )
            * torch.sin(alpha)
        )

        r22 = (
            Ei
            / Ef
            * (
                torch.cos(alpha)
                + torch.sqrt(2 / eta) * torch.cos(phi) * torch.sin(alpha)
            )
        )

        r56 = torch.tensor(0.0)
        beta0 = torch.tensor(1.0)
        beta1 = torch.tensor(1.0)

        k = 2 * torch.pi * self.frequency / torch.tensor(constants.speed_of_light)
        r55_cor = 0
        if self.voltage != 0 and energy != 0:  # TODO: Do we need this if?
            beta0 = torch.sqrt(1 - 1 / Ei**2)
            beta1 = torch.sqrt(1 - 1 / Ef**2)

            r56 = -self.length / (Ef**2 * Ei * beta1) * (Ef + Ei) / (beta1 + beta0)
            g0 = Ei
            g1 = Ef
            r55_cor = (
                k
                * self.length
                * beta0
                * self.voltage
                / electron_mass_eV
                * torch.sin(phi)
                * (g0 * g1 * (beta0 * beta1 - 1) + 1)
                / (beta1 * g1 * (g0 - g1) ** 2)
            )

        r66 = Ei / Ef * beta0 / beta1
        r65 = k * torch.sin(phi) * self.voltage / (Ef * beta1 * electron_mass_eV)

        R = torch.eye(7, device=self.device)
        R[0, 0] = r11
        R[0, 1] = r12
        R[1, 0] = r21
        R[1, 1] = r22
        R[2, 2] = r11
        R[2, 3] = r12
        R[3, 2] = r21
        R[3, 3] = r22
        R[4, 4] = 1 + r55_cor
        R[4, 5] = r56
        R[5, 4] = r65
        R[5, 5] = r66

        return R

    def split(self, resolution: torch.Tensor) -> list[Element]:
        # TODO: Implement splitting for cavity properly, for now just returns the
        # element itself
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.4

        patch = Rectangle(
            (s, 0), self.length, height, color="gold", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length", "voltage", "phase", "frequency"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"voltage={repr(self.voltage)}, "
            + f"phase={repr(self.voltage)}, "
            + f"frequency={repr(self.frequency)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class BPM(Element):
    """
    Beam Position Monitor (BPM) in a particle accelerator.

    :param is_active: If `True` the BPM is active and will record the beam's position.
        If `False` the BPM is inactive and will not record the beam's position.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self, is_active: bool = False, name: Optional[str] = None, device: str = "auto"
    ) -> None:
        super().__init__(name=name, device=device)

        self.is_active = is_active
        self.reading = None

    @property
    def is_skippable(self) -> bool:
        return not self.is_active

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        return torch.eye(7, device=self.device)

    def track(self, incoming: Beam) -> Beam:
        if incoming is Beam.empty:
            self.reading = None
        elif isinstance(incoming, ParameterBeam):
            self.reading = torch.stack([incoming._mu_x, incoming._mu_y])
        elif isinstance(incoming, ParticleBeam):
            self.reading = torch.stack([incoming.mu_x, incoming.mu_y])
        else:
            raise TypeError(f"Parameter incoming is of invalid type {type(incoming)}")

        return deepcopy(incoming)

    def split(self, resolution: torch.Tensor) -> list[Element]:
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        patch = Rectangle(
            (s, -0.3), 0, 0.3 * 2, color="darkkhaki", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={repr(self.name)},"
            f" device={repr(self.device)})"
        )


class Marker(Element):
    """
    General Marker / Monitor element

    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(self, name: Optional[str] = None, device: str = "auto") -> None:
        super().__init__(name=name, device=device)

    def transfer_map(self, energy):
        return torch.eye(7, device=self.device)

    def track(self, incoming):
        return incoming

    @property
    def is_skippable(self) -> bool:
        return True

    def split(self, resolution: torch.Tensor) -> list[Element]:
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        # Do nothing on purpose. Maybe later we decide markers should be shown, but for
        # now they are invisible.
        pass

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={repr(self.name)},"
            f" device={repr(self.device)})"
        )


class Screen(Element):
    """
    Diagnostic screen in a particle accelerator.

    :param resolution: Resolution of the camera sensor looking at the screen given as
        Tensor `(width, height)`.
    :param pixel_size: Size of a pixel on the screen in meters given as a Tensor
        `(width, height)`.
    :param binning: Binning used by the camera.
    :param misalignment: Misalignment of the screen in meters given as a Tensor
        `(x, y)`.
    :param is_active: If `True` the screen is active and will record the beam's
        distribution. If `False` the screen is inactive and will not record the beam's
        distribution.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        resolution: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        pixel_size: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        binning: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        misalignment: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        is_active: bool = False,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.resolution = (
            resolution if resolution is not None else torch.tensor((1024, 1024))
        )
        self.pixel_size = (
            pixel_size if pixel_size is not None else torch.tensor((1e-3, 1e-3))
        )
        self.binning = binning if binning is not None else torch.tensor(1)
        self.misalignment = (
            misalignment if misalignment is not None else torch.tensor((0.0, 0.0))
        )
        self.is_active = is_active

        self.set_read_beam(None)
        self.cached_reading = None

    @property
    def is_skippable(self) -> bool:
        return not self.is_active

    @property
    def effective_resolution(self) -> torch.Tensor:
        return self.resolution / self.binning

    @property
    def effective_pixel_size(self) -> torch.Tensor:
        return self.pixel_size * self.binning

    @property
    def extent(self) -> tuple[float, float, float, float]:
        return (
            -self.resolution[0] * self.pixel_size[0] / 2,
            self.resolution[0] * self.pixel_size[0] / 2,
            -self.resolution[1] * self.pixel_size[1] / 2,
            self.resolution[1] * self.pixel_size[1] / 2,
        )

    @property
    def pixel_bin_edges(self) -> tuple[torch.Tensor, torch.Tensor]:
        return (
            torch.linspace(
                -self.resolution[0] * self.pixel_size[0] / 2,
                self.resolution[0] * self.pixel_size[0] / 2,
                int(self.effective_resolution[0]) + 1,
            ),
            torch.linspace(
                -self.resolution[1] * self.pixel_size[1] / 2,
                self.resolution[1] * self.pixel_size[1] / 2,
                int(self.effective_resolution[1]) + 1,
            ),
        )

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        return torch.eye(7, device=self.device)

    def track(self, incoming: Beam) -> Beam:
        if self.is_active:
            copy_of_incoming = deepcopy(incoming)

            if isinstance(incoming, ParameterBeam):
                copy_of_incoming._mu[0] -= self.misalignment[0]
                copy_of_incoming._mu[2] -= self.misalignment[1]
            elif isinstance(incoming, ParticleBeam):
                copy_of_incoming.particles[:, 0] -= self.misalignment[0]
                copy_of_incoming.particles[:, 1] -= self.misalignment[1]

            self.set_read_beam(copy_of_incoming)

            return Beam.empty
        else:
            return incoming

    @property
    def reading(self) -> torch.Tensor:
        if self.cached_reading is not None:
            return self.cached_reading

        read_beam = self.get_read_beam()
        if read_beam is Beam.empty or read_beam is None:
            image = torch.zeros(
                (self.effective_resolution[1], self.effective_resolution[0])
            )
        elif isinstance(read_beam, ParameterBeam):
            transverse_mu = torch.stack([read_beam._mu[0], read_beam._mu[2]])
            transverse_cov = torch.stack(
                [
                    torch.stack([read_beam._cov[0, 0], read_beam._cov[0, 2]]),
                    torch.stack([read_beam._cov[2, 0], read_beam._cov[2, 2]]),
                ]
            )
            dist = multivariate_normal(
                mean=transverse_mu, cov=transverse_cov, allow_singular=True
            )

            left = self.extent[0]
            right = self.extent[1]
            hstep = self.pixel_size[0] * self.binning
            bottom = self.extent[2]
            top = self.extent[3]
            vstep = self.pixel_size[1] * self.binning
            x, y = torch.mgrid[left:right:hstep, bottom:top:vstep]
            pos = torch.dstack((x, y))
            image = dist.pdf(pos)
            image = torch.flipud(image.T)
        elif isinstance(read_beam, ParticleBeam):
            image, _ = torch.histogramdd(
                torch.stack((read_beam.xs, read_beam.ys)).T,
                bins=self.pixel_bin_edges,
            )
            image = torch.flipud(image.T)
            image = image.cpu()
        else:
            raise TypeError(f"Read beam is of invalid type {type(read_beam)}")

        self.cached_reading = image
        return image

    def get_read_beam(self) -> Beam:
        # Using these get and set methods instead of Python's property decorator to
        # prevent `nn.Module` from intercepting the read beam, which is itself an
        # `nn.Module`, and registering it as a submodule of the screen.
        return self._read_beam[0] if self._read_beam is not None else None

    def set_read_beam(self, value: Beam) -> None:
        # Using these get and set methods instead of Python's property decorator to
        # prevent `nn.Module` from intercepting the read beam, which is itself an
        # `nn.Module`, and registering it as a submodule of the screen.
        self._read_beam = [value]
        self.cached_reading = None

    def split(self, resolution: torch.Tensor) -> list[Element]:
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        patch = Rectangle(
            (s, -0.6), 0, 0.6 * 2, color="tab:green", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + [
            "resolution",
            "pixel_size",
            "binning",
            "misalignment",
            "is_active",
        ]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(resolution={repr(self.resolution)}, "
            + f"pixel_size={repr(self.pixel_size)}, "
            + f"binning={repr(self.binning)}, "
            + f"misalignment={repr(self.misalignment)}, "
            + f"is_active={repr(self.is_active)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class Aperture(Element):
    """
    Physical aperture.

    :param x_max: half size horizontal offset in [m]
    :param y_max: half size vertical offset in [m]
    :param shape: Shape of the aperture. Can be "rectangular" or "elliptical".
    :param is_active: If the aperture actually blocks particles.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        x_max: Union[torch.Tensor, nn.Parameter] = torch.inf,
        y_max: Union[torch.Tensor, nn.Parameter] = torch.inf,
        shape: Literal["rectangular", "elliptical"] = "rectangular",
        is_active: bool = True,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.x_max = x_max
        self.y_max = y_max
        self.shape = shape
        self.is_active = is_active

        self.lost_particles = None

    @property
    def is_skippable(self) -> bool:
        return not self.is_active

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        return torch.eye(7, device=self.device)

    def track(self, incoming: Beam) -> Beam:
        # Only apply aperture to particle beams and if the element is active
        if not (isinstance(incoming, ParticleBeam) and self.is_active):
            return incoming

        assert self.x_max >= 0 and self.y_max >= 0
        assert self.shape in [
            "rectangular",
            "elliptical",
        ], f"Unknown aperture shape {self.shape}"

        if self.shape == "rectangular":
            survived_mask = torch.logical_and(
                torch.logical_and(incoming.xs > -self.x_max, incoming.xs < self.x_max),
                torch.logical_and(incoming.ys > -self.y_max, incoming.ys < self.y_max),
            )
        elif self.shape == "elliptical":
            survived_mask = (
                incoming.xs**2 / self.x_max**2 + incoming.ys**2 / self.y_max**2
            ) <= 1.0
        outgoing_particles = incoming.particles[survived_mask]

        self.lost_particles = incoming.particles[torch.logical_not(survived_mask)]

        return (
            ParticleBeam(outgoing_particles, incoming.energy, device=incoming.device)
            if outgoing_particles.shape[0] > 0
            else ParticleBeam.empty
        )

    def split(self, resolution: torch.Tensor) -> list[Element]:
        # TODO: Implement splitting for aperture properly, for now just return self
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.4

        dummy_length = 0.0

        patch = Rectangle(
            (s, 0), dummy_length, height, color="tab:pink", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + [
            "x_max",
            "y_max",
            "shape",
            "is_active",
        ]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x_max={repr(self.x_max)}, "
            + f"y_max={repr(self.y_max)}, "
            + f"shape={repr(self.shape)}, "
            + f"is_active={repr(self.is_active)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class Undulator(Element):
    """
    Element representing an undulator in a particle accelerator.

    NOTE Currently behaves like a drift section but is plotted distinctively.

    :param length: Length in meters.
    :param is_active: Indicates if the undulator is active or not. Currently has no
        effect.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Union[torch.Tensor, nn.Parameter],
        is_active: bool = False,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length
        self.is_active = is_active

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        gamma = energy / rest_energy
        igamma2 = 1 / gamma**2 if gamma != 0 else torch.tensor(0.0)

        tm = torch.eye(7, device=self.device)
        tm[0, 1] = self.length
        tm[2, 3] = self.length
        tm[4, 5] = self.length * igamma2

        return tm

    @property
    def is_skippable(self) -> bool:
        return True

    def split(self, resolution: torch.Tensor) -> list[Element]:
        # TODO: Implement splitting for undulator properly, for now just return self
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.4

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:purple", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"is_active={repr(self.is_active)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class Solenoid(Element):
    """
    Solenoid magnet.

    Implemented according to A.W.Chao P74

    :param length: Length in meters.
    :param k: Normalised strength of the solenoid magnet B0/(2*Brho). B0 is the field
        inside the solenoid, Brho is the momentum of central trajectory.
    :param misalignment: Misalignment vector of the solenoid magnet in x- and
        y-directions.
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self,
        length: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        k: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        misalignment: Optional[Union[torch.Tensor, nn.Parameter]] = None,
        name: Optional[str] = None,
        device: str = "auto",
    ) -> None:
        super().__init__(name=name, device=device)

        self.length = length if length is not None else torch.tensor(0.0)
        self.k = k if k is not None else torch.tensor(0.0)
        self.misalignment = (
            misalignment if misalignment is not None else torch.tensor((0.0, 0.0))
        )

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        gamma = energy / rest_energy
        c = torch.cos(self.length * self.k)
        s = torch.sin(self.length * self.k)
        if self.k == 0:
            s_k = self.length
        else:
            s_k = s / self.k
        r56 = torch.tensor(0.0)
        if gamma != 0:
            gamma2 = gamma * gamma
            beta = torch.sqrt(1.0 - 1.0 / gamma2)
            r56 -= self.length / (beta * beta * gamma2)

        R = torch.eye(7, device=self.device)
        R[0, 0] = c**2
        R[0, 1] = c * s_k
        R[0, 2] = s * c
        R[0, 3] = s * s_k
        R[1, 0] = -self.k * s * c
        R[1, 1] = c**2
        R[1, 2] = -self.k * s**2
        R[1, 3] = s * c
        R[2, 0] = -s * c
        R[2, 1] = -s * s_k
        R[2, 2] = c**2
        R[2, 3] = c * s_k
        R[3, 0] = self.k * s**2
        R[3, 1] = -s * c
        R[3, 2] = -self.k * s * c
        R[3, 3] = c**2
        R[4, 5] = r56

        R = R.real

        if self.misalignment[0] == 0 and self.misalignment[1] == 0:
            return R
        else:
            R_exit, R_entry = misalignment_matrix(self.misalignment, self.device)
            R = torch.matmul(R_exit, torch.matmul(R, R_entry))
            return R

    @property
    def is_active(self) -> bool:
        return self.k != 0

    def is_skippable(self) -> bool:
        return True

    def split(self, resolution: torch.Tensor) -> list[Element]:
        # TODO: Implement splitting for solenoid properly, for now just return self
        return [self]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        alpha = 1 if self.is_active else 0.2
        height = 0.8

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:orange", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["length", "k", "misalignment"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={repr(self.length)}, "
            + f"k={repr(self.k)}, "
            + f"misalignment={repr(self.misalignment)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )


class Segment(Element):
    """
    Segment of a particle accelerator consisting of several elements.

    :param cell: List of Cheetah elements that describe an accelerator (section).
    :param name: Unique identifier of the element.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self, elements: list[Element], name: str = "unnamed", device: str = "auto"
    ) -> None:
        super().__init__(name=name, device=device)

        self.elements = nn.ModuleList(elements)

        for element in self.elements:
            element.device = self.device

            # Make elements accessible via .name attribute. If multiple elements have
            # the same name, they are accessible via a list.
            if element.name in self.__dict__:
                if isinstance(self.__dict__[element.name], list):
                    self.__dict__[element.name].append(element)
                else:  # Is instance of cheetah.Element
                    self.__dict__[element.name] = [self.__dict__[element.name], element]
            else:
                self.__dict__[element.name] = element

    def subcell(self, start: str, end: str, **kwargs) -> "Segment":
        """Extract a subcell `[start, end]` from an this segment."""
        subcell = []
        is_in_subcell = False
        for element in self.elements:
            if element.name == start:
                is_in_subcell = True
            if is_in_subcell:
                subcell.append(element)
            if element.name == end:
                break

        return self.__class__(subcell, device=self.device, **kwargs)

    def flattened(self) -> "Segment":
        """
        Return a flattened version of the segment, i.e. one where all subsegments are
        resolved and their elements entered into a top-level segment.
        """
        flattened_elements = []
        for element in self.elements:
            if isinstance(element, Segment):
                flattened_elements += element.flattened().elements
            else:
                flattened_elements.append(element)

        return Segment(elements=flattened_elements, name=self.name, device=self.device)

    @classmethod
    def from_lattice_json(cls, filepath: str) -> "Segment":
        """
        Load a Cheetah model from a JSON file.

        :param filename: Name/path of the file to load the lattice from.
        :return: Loaded Cheetah `Segment`.
        """
        return load_cheetah_model(filepath)

    def to_lattice_json(
        self,
        filepath: str,
        title: Optional[str] = None,
        info: str = "This is a placeholder lattice description",
    ) -> None:
        """
        Save a Cheetah model to a JSON file.

        :param filename: Name/path of the file to save the lattice to.
        :param title: Title of the lattice. If not provided, defaults to the name of the
        `Segment` object. If that also does not have a name, defaults to "Unnamed
            Lattice".
        :param info: Information about the lattice. Defaults to "This is a placeholder
            lattice description".
        """
        save_cheetah_model(self, filepath, title, info)

    @classmethod
    def from_ocelot(
        cls, cell, name: Optional[str] = None, warnings: bool = True, **kwargs
    ) -> "Segment":
        """
        Translate an Ocelot cell to a Cheetah `Segment`.

        NOTE Objects not supported by Cheetah are translated to drift sections. Screen
        objects are created only from `ocelot.Monitor` objects when the string "BSC" is
        contained in their `id` attribute. Their screen properties are always set to
        default values and most likely need adjusting afterwards. BPM objects are only
        created from `ocelot.Monitor` objects when their id has a substring "BPM".

        :param cell: Ocelot cell, i.e. a list of Ocelot elements to be converted.
        :param name: Unique identifier for the entire segment.
        :param warnings: Whether to print warnings when objects are not supported by
            Cheetah or converted with potentially unexpected behavior.
        :return: Cheetah segment closely resembling the Ocelot cell.
        """
        from cheetah.converters.nocelot import ocelot2cheetah

        converted = [ocelot2cheetah(element, warnings=warnings) for element in cell]
        return cls(converted, name=name, **kwargs)

    @classmethod
    def from_bmad(
        cls, bmad_lattice_file_path: str, environment_variables: Optional[dict] = None
    ) -> "Segment":
        """
        Read a Cheetah segment from a Bmad lattice file.

        NOTE: This function was designed at the example of the LCLS lattice. While this
        lattice is extensive, this function might not properly convert all features of
        a Bmad lattice. If you find that this function does not work for your lattice,
        please open an issue on GitHub.

        :param bmad_lattice_file_path: Path to the Bmad lattice file.
        :param environment_variables: Dictionary of environment variables to use when
            parsing the lattice file.
        :return: Cheetah `Segment` representing the Bmad lattice.
        """
        bmad_lattice_file_path = Path(bmad_lattice_file_path)
        return convert_bmad_lattice(bmad_lattice_file_path, environment_variables)

    @property
    def is_skippable(self) -> bool:
        return all(element.is_skippable for element in self.elements)

    @property
    def length(self) -> float:
        lengths = torch.stack(
            [element.length for element in self.elements if hasattr(element, "length")]
        )
        return torch.sum(lengths)

    def transfer_map(self, energy: torch.Tensor) -> torch.Tensor:
        if self.is_skippable:
            tm = torch.eye(7, dtype=torch.float32, device=self.device)
            for element in self.elements:
                tm = torch.matmul(element.transfer_map(energy), tm)
            return tm
        else:
            return None

    def track(self, incoming: Beam) -> Beam:
        if self.is_skippable:
            return super().track(incoming)
        else:
            todos = []
            for element in self.elements:
                if not element.is_skippable:
                    todos.append(element)
                elif not todos or not todos[-1].is_skippable:
                    todos.append(Segment([element], device=self.device))
                else:
                    todos[-1].elements.append(element)

            for todo in todos:
                incoming = todo(incoming)

            return incoming

    def split(self, resolution: torch.Tensor) -> list[Element]:
        return [
            split_element
            for element in self.elements
            for split_element in element.split(resolution)
        ]

    def plot(self, ax: matplotlib.axes.Axes, s: float) -> None:
        element_lengths = [
            element.length if hasattr(element, "length") else 0.0
            for element in self.elements
        ]
        element_ss = [0] + [
            sum(element_lengths[: i + 1]) for i, _ in enumerate(element_lengths)
        ]
        element_ss = [s + element_s for element_s in element_ss]

        ax.plot([0, element_ss[-1]], [0, 0], "--", color="black")

        for element, s in zip(self.elements, element_ss[:-1]):
            element.plot(ax, s)

        ax.set_ylim(-1, 1)
        ax.set_xlabel("s (m)")
        ax.set_yticks([])

    def plot_reference_particle_traces(
        self,
        axx: matplotlib.axes.Axes,
        axy: matplotlib.axes.Axes,
        beam: Optional[Beam] = None,
        num_particles: int = 10,
        resolution: float = 0.01,
    ) -> None:
        """
        Plot `n` reference particles along the segment view in x- and y-direction.

        :param axx: Axes to plot the particle traces into viewed in x-direction.
        :param axy: Axes to plot the particle traces into viewed in y-direction.
        :param beam: Entering beam from which the reference particles are sampled.
        :param num_particles: Number of reference particles to plot. Must not be larger
            than number of particles passed in `beam`.
        :param resolution: Minimum resolution of the tracking of the reference particles
            in the plot.
        """
        reference_segment = deepcopy(self)
        splits = reference_segment.split(resolution=torch.tensor(resolution))

        split_lengths = [
            split.length if hasattr(split, "length") else 0.0 for split in splits
        ]
        ss = [0] + [sum(split_lengths[: i + 1]) for i, _ in enumerate(split_lengths)]

        references = []
        if beam is None:
            initial = ParticleBeam.make_linspaced(
                num_particles=num_particles, device="cpu"
            )
            references.append(initial)
        else:
            initial = ParticleBeam.make_linspaced(
                num_particles=num_particles,
                mu_x=beam.mu_x,
                mu_xp=beam.mu_xp,
                mu_y=beam.mu_y,
                mu_yp=beam.mu_yp,
                sigma_x=beam.sigma_x,
                sigma_xp=beam.sigma_xp,
                sigma_y=beam.sigma_y,
                sigma_yp=beam.sigma_yp,
                sigma_s=beam.sigma_s,
                sigma_p=beam.sigma_p,
                energy=beam.energy,
                device="cpu",
            )
            references.append(initial)
        for split in splits:
            sample = split(references[-1])
            references.append(sample)

        for particle_index in range(num_particles):
            xs = [
                float(reference_beam.xs[particle_index].cpu())
                for reference_beam in references
                if reference_beam is not Beam.empty
            ]
            axx.plot(ss[: len(xs)], xs)
        axx.set_xlabel("s (m)")
        axx.set_ylabel("x (m)")
        axx.grid()

        for particle_index in range(num_particles):
            ys = [
                float(reference_beam.ys[particle_index].cpu())
                for reference_beam in references
                if reference_beam is not Beam.empty
            ]
            axy.plot(ss[: len(ys)], ys)
        axx.set_xlabel("s (m)")
        axy.set_ylabel("y (m)")
        axy.grid()

    def plot_overview(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        beam: Optional[Beam] = None,
        n: int = 10,
        resolution: float = 0.01,
    ) -> None:
        """
        Plot an overview of the segment with the lattice and traced reference particles.

        :param fig: Figure to plot the overview into.
        :param beam: Entering beam from which the reference particles are sampled.
        :param n: Number of reference particles to plot. Must not be larger than number
            of particles passed in `beam`.
        :param resolution: Minimum resolution of the tracking of the reference particles
            in the plot.
        """
        if fig is None:
            fig = plt.figure()
        gs = fig.add_gridspec(3, hspace=0, height_ratios=[2, 2, 1])
        axs = gs.subplots(sharex=True)

        axs[0].set_title("Reference Particle Traces")
        self.plot_reference_particle_traces(axs[0], axs[1], beam, n, resolution)

        self.plot(axs[2], 0)

        plt.tight_layout()

    def plot_twiss(self, beam: Beam, ax: Optional[Any] = None) -> None:
        """Plot twiss parameters along the segment."""
        longitudinal_beams = [beam]
        s_positions = [0.0]
        for element in self.elements:
            if not hasattr(element, "length") or element.length == 0:
                continue

            outgoing = element.track(longitudinal_beams[-1])

            longitudinal_beams.append(outgoing)
            s_positions.append(s_positions[-1] + element.length)

        beta_x = [beam.beta_x for beam in longitudinal_beams]
        beta_y = [beam.beta_y for beam in longitudinal_beams]

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        ax.set_title("Twiss Parameters")
        ax.set_xlabel("s (m)")
        ax.set_ylabel(r"$\beta$ (m)")

        ax.plot(s_positions, beta_x, label=r"$\beta_x$", c="tab:red")
        ax.plot(s_positions, beta_y, label=r"$\beta_y$", c="tab:green")

        ax.legend()
        plt.tight_layout()

    @property
    def defining_features(self) -> list[str]:
        return super().defining_features + ["elements"]

    def plot_twiss_over_lattice(self, beam: Beam, figsize=(8, 4)) -> None:
        """Plot twiss parameters in a plot over a plot of the lattice."""
        fig = plt.figure(figsize=figsize)
        gs = fig.add_gridspec(2, hspace=0, height_ratios=[3, 1])
        axs = gs.subplots(sharex=True)

        self.plot_twiss(beam, ax=axs[0])
        self.plot(axs[1], 0)

        plt.tight_layout()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(elements={repr(self.elements)}, "
            + f"name={repr(self.name)}, "
            + f"device={repr(self.device)})"
        )



import numpy as np
import torch

from cheetah import Drift, ParameterBeam, Quadrupole


def test_quadrupole_off():
    """
    Test that a quadrupole with k1=0 behaves still like a drift.
    """
    quadrupole = Quadrupole(length=torch.tensor(1.0), k1=torch.tensor(0.0))
    drift = Drift(length=torch.tensor(1.0))
    incoming_beam = ParameterBeam.from_parameters(
        sigma_xp=torch.tensor(2e-7), sigma_yp=torch.tensor(2e-7)
    )
    outbeam_quad = quadrupole(incoming_beam)
    outbeam_drift = drift(incoming_beam)

    quadrupole.k1 = torch.tensor(1.0)
    outbeam_quad_on = quadrupole(incoming_beam)

    assert np.allclose(outbeam_quad.sigma_x, outbeam_drift.sigma_x)
    assert not np.allclose(outbeam_quad_on.sigma_x, outbeam_drift.sigma_x)



import pytest
import torch

from cheetah.accelerator import Drift, Quadrupole, Segment


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_drift_equal():
    """Test that equal drift sections are recognised as such."""
    drift_1 = Drift(name="drift_eq", length=torch.tensor(0.1))
    drift_2 = Drift(name="drift_eq", length=torch.tensor(0.1))
    assert drift_1 == drift_2


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_drift_not_equal():
    """Test that unequal drift sections are recognised as such."""
    drift_1 = Drift(name="drift_1", length=torch.tensor(0.1))
    drift_2 = Drift(name="drift_2", length=torch.tensor(0.2))
    assert drift_1 != drift_2


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_quadrupole_equal():
    """Test that equal quadrupoles are recognised as such."""
    quadrupole_1 = Quadrupole(
        name="quadrupole_eq", length=torch.tensor(0.1), k1=torch.tensor(0.2)
    )
    quadrupole_2 = Quadrupole(
        name="quadrupole_eq", length=torch.tensor(0.1), k1=torch.tensor(0.2)
    )
    assert quadrupole_1 == quadrupole_2


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_quadrupole_not_equal():
    """Test that unequal quadrupoles are recognised as such."""
    quadrupole_1 = Quadrupole(
        name="quadrupole_1", length=torch.tensor(0.1), k1=torch.tensor(0.2)
    )
    quadrupole_2 = Quadrupole(
        name="quadrupole_2", length=torch.tensor(0.2), k1=torch.tensor(0.2)
    )
    assert quadrupole_1 != quadrupole_2


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_segment_equal():
    """Test that equal segments are recognised as such."""
    segment_1 = Segment(
        elements=[
            Drift(name="d", length=torch.tensor(0.1)),
            Quadrupole(name="q", length=torch.tensor(0.1), k1=torch.tensor(0.2)),
        ],
        name="test_segment",
    )
    segment_2 = Segment(
        elements=[
            Drift(name="d", length=torch.tensor(0.1)),
            Quadrupole(name="q", length=torch.tensor(0.1), k1=torch.tensor(0.2)),
        ],
        name="test_segment",
    )
    assert segment_1 == segment_2


@pytest.mark.skip(reason="Equality is broken since making elements nn.Modules.")
def test_segment_not_equal():
    """Test that unequal segments are recognised as such."""
    segment_1 = Segment(
        elements=[
            Drift(name="d", length=torch.tensor(0.1)),
            Quadrupole(name="q", length=torch.tensor(0.1), k1=torch.tensor(0.2)),
        ],
        name="test_segment",
    )
    segment_2 = Segment(
        elements=[
            Quadrupole(name="q", length=torch.tensor(0.2), k1=torch.tensor(0.2)),
            Drift(name="d", length=torch.tensor(0.1)),
        ],
        name="test_segment",
    )
    assert segment_1 != segment_2



from typing import Optional

import numpy as np
import torch
from scipy.constants import physical_constants
from torch import nn
from torch.distributions import MultivariateNormal

electron_mass_eV = torch.tensor(
    physical_constants["electron mass energy equivalent in MeV"][0] * 1e6
)


class Beam(nn.Module):
    empty = "I'm an empty beam!"

    @classmethod
    def from_parameters(
        cls,
        mu_x: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        cor_x: Optional[torch.Tensor] = None,
        cor_y: Optional[torch.Tensor] = None,
        cor_s: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
    ) -> "Beam":
        """
        Create beam that with given beam parameters.

        :param n: Number of particles to generate.
        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_xp: Center of the particle distribution on x'=px/px'
            (trace space) in rad.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_yp: Center of the particle distribution on y' in rad.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_xp: Sigma of the particle distribution in x' direction in rad.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_yp: Sigma of the particle distribution in y' direction in rad.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param energy: Energy of the beam in eV.
        """
        raise NotImplementedError

    @classmethod
    def from_twiss(
        cls,
        beta_x: Optional[torch.Tensor] = None,
        alpha_x: Optional[torch.Tensor] = None,
        emittance_x: Optional[torch.Tensor] = None,
        beta_y: Optional[torch.Tensor] = None,
        alpha_y: Optional[torch.Tensor] = None,
        emittance_y: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
    ) -> "Beam":
        """
        Create a beam from twiss parameters.

        :param beta_x: Beta function in x direction in meters.
        :param alpha_x: Alpha function in x direction in meters.
        :param emittance_x: Emittance in x direction.
        :param beta_y: Beta function in y direction in meters.
        :param alpha_y: Alpha function in y direction in meters.
        :param emittance_y: Emittance in y direction.
        :param energy: Energy of the beam in eV.
        """
        raise NotImplementedError

    @classmethod
    def from_ocelot(cls, parray) -> "Beam":
        """
        Convert an Ocelot ParticleArray `parray` to a Cheetah Beam.
        """
        raise NotImplementedError

    @classmethod
    def from_astra(cls, path: str, **kwargs) -> "Beam":
        """Load an Astra particle distribution as a Cheetah Beam."""
        raise NotImplementedError

    def transformed_to(
        self,
        mu_x: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
    ) -> "Beam":
        """
        Create version of this beam that is transformed to new beam parameters.

        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_xp: Center of the particle distribution on px in meters.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_yp: Center of the particle distribution on py in meters.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_xp: Sigma of the particle distribution in px direction in meters.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_yp: Sigma of the particle distribution in py direction in meters.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param energy: Energy of the beam in eV.
        """
        mu_x = mu_x if mu_x is not None else self.mu_x
        mu_xp = mu_xp if mu_xp is not None else self.mu_xp
        mu_y = mu_y if mu_y is not None else self.mu_y
        mu_yp = mu_yp if mu_yp is not None else self.mu_yp
        sigma_x = sigma_x if sigma_x is not None else self.sigma_x
        sigma_xp = sigma_xp if sigma_xp is not None else self.sigma_xp
        sigma_y = sigma_y if sigma_y is not None else self.sigma_y
        sigma_yp = sigma_yp if sigma_yp is not None else self.sigma_yp
        sigma_s = sigma_s if sigma_s is not None else self.sigma_s
        sigma_p = sigma_p if sigma_p is not None else self.sigma_p
        energy = energy if energy is not None else self.energy

        return self.__class__.from_parameters(
            mu_x=mu_x,
            mu_xp=mu_xp,
            mu_y=mu_y,
            mu_yp=mu_yp,
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
        )

    @property
    def parameters(self) -> dict:
        return {
            "mu_x": self.mu_x,
            "mu_xp": self.mu_xp,
            "mu_y": self.mu_y,
            "mu_yp": self.mu_yp,
            "sigma_x": self.sigma_x,
            "sigma_xp": self.sigma_xp,
            "sigma_y": self.sigma_y,
            "sigma_yp": self.sigma_yp,
            "sigma_s": self.sigma_s,
            "sigma_p": self.sigma_p,
            "energy": self.energy,
        }

    @property
    def mu_x(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_x(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mu_xp(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_xp(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mu_y(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_y(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mu_yp(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_yp(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mu_s(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_s(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mu_p(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def sigma_p(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def relativistic_gamma(self) -> torch.Tensor:
        return self.energy / electron_mass_eV

    @property
    def relativistic_beta(self) -> torch.Tensor:
        relativistic_beta = (
            torch.sqrt(1 - 1 / (self.relativistic_gamma**2))
            if torch.abs(self.relativistic_gamma) > 0
            else torch.tensor(1.0)
        )
        return relativistic_beta

    @property
    def sigma_xxp(self) -> torch.Tensor:
        # the covariance of (x,x') ~ $\sigma_{xx'}$
        raise NotImplementedError

    @property
    def sigma_yyp(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def emittance_x(self) -> torch.Tensor:
        """Emittance of the beam in x direction in m*rad."""
        return torch.sqrt(self.sigma_x**2 * self.sigma_xp**2 - self.sigma_xxp**2)

    @property
    def normalized_emittance_x(self) -> torch.Tensor:
        """Normalized emittance of the beam in x direction in m*rad."""
        return self.emittance_x * self.relativistic_beta * self.relativistic_gamma

    @property
    def beta_x(self) -> torch.Tensor:
        """Beta function in x direction in meters."""
        return self.sigma_x**2 / self.emittance_x

    @property
    def alpha_x(self) -> torch.Tensor:
        return -self.sigma_xxp / self.emittance_x

    @property
    def emittance_y(self) -> torch.Tensor:
        """Emittance of the beam in y direction in m*rad."""
        return torch.sqrt(self.sigma_y**2 * self.sigma_yp**2 - self.sigma_yyp**2)

    @property
    def normalized_emittance_y(self) -> torch.Tensor:
        """Normalized emittance of the beam in y direction in m*rad."""
        return self.emittance_y * self.relativistic_beta * self.relativistic_gamma

    @property
    def beta_y(self) -> torch.Tensor:
        """Beta function in y direction in meters."""
        return self.sigma_y**2 / self.emittance_y

    @property
    def alpha_y(self) -> torch.Tensor:
        return -self.sigma_yyp / self.emittance_y

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(mu_x={self.mu_x}, mu_xp={self.mu_xp},"
            f" mu_y={self.mu_y}, mu_yp={self.mu_yp}, sigma_x={self.sigma_x},"
            f" sigma_xp={self.sigma_xp}, sigma_y={self.sigma_y},"
            f" sigma_yp={self.sigma_yp}, sigma_s={self.sigma_s},"
            f" sigma_p={self.sigma_p}, energy={self.energy})"
        )


class ParameterBeam(Beam):
    """
    Beam of charged particles, where each particle is simulated.

    :param mu: Mu vector of the beam.
    :param cov: Covariance matrix of the beam.
    :param energy: Energy of the beam in eV.
    :param device: Device to use for the beam. If "auto", use CUDA if available.
        Note: Compuationally it would be faster to use CPU for ParameterBeam.
    """

    def __init__(
        self,
        mu: torch.Tensor,
        cov: torch.Tensor,
        energy: torch.Tensor,
        device: str = "auto",
    ) -> None:
        super().__init__()

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

        self._mu = mu.to(device)
        self._cov = cov.to(device)
        self.energy = energy

    @classmethod
    def from_parameters(
        cls,
        mu_x: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        cor_x: Optional[torch.Tensor] = None,
        cor_y: Optional[torch.Tensor] = None,
        cor_s: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
        device: str = "auto",
    ) -> "ParameterBeam":
        # Set default values without function call in function signature
        mu_x = mu_x if mu_x is not None else torch.tensor(0.0)
        mu_xp = mu_xp if mu_xp is not None else torch.tensor(0.0)
        mu_y = mu_y if mu_y is not None else torch.tensor(0.0)
        mu_yp = mu_yp if mu_yp is not None else torch.tensor(0.0)
        sigma_x = sigma_x if sigma_x is not None else torch.tensor(175e-9)
        sigma_xp = sigma_xp if sigma_xp is not None else torch.tensor(2e-7)
        sigma_y = sigma_y if sigma_y is not None else torch.tensor(175e-9)
        sigma_yp = sigma_yp if sigma_yp is not None else torch.tensor(2e-7)
        sigma_s = sigma_s if sigma_s is not None else torch.tensor(1e-6)
        sigma_p = sigma_p if sigma_p is not None else torch.tensor(1e-6)
        cor_x = cor_x if cor_x is not None else torch.tensor(0.0)
        cor_y = cor_y if cor_y is not None else torch.tensor(0.0)
        cor_s = cor_s if cor_s is not None else torch.tensor(0.0)
        energy = energy if energy is not None else torch.tensor(1e8)

        mu = torch.stack(
            [
                mu_x,
                mu_xp,
                mu_y,
                mu_yp,
                torch.tensor(0.0),
                torch.tensor(0.0),
                torch.tensor(1.0),
            ]
        )

        cov = torch.zeros(7, 7)
        cov[0, 0] = sigma_x**2
        cov[0, 1] = cor_x
        cov[1, 0] = cor_x
        cov[1, 1] = sigma_xp**2
        cov[2, 2] = sigma_y**2
        cov[2, 3] = cor_y
        cov[3, 2] = cor_y
        cov[3, 3] = sigma_yp**2
        cov[4, 4] = sigma_s**2
        cov[4, 5] = cor_s
        cov[5, 4] = cor_s
        cov[5, 5] = sigma_p**2

        return cls(mu=mu, cov=cov, energy=energy, device=device)

    @classmethod
    def from_twiss(
        cls,
        beta_x: Optional[torch.Tensor] = None,
        alpha_x: Optional[torch.Tensor] = None,
        emittance_x: Optional[torch.Tensor] = None,
        beta_y: Optional[torch.Tensor] = None,
        alpha_y: Optional[torch.Tensor] = None,
        emittance_y: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        cor_s: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
        device: str = "auto",
    ) -> "ParameterBeam":
        # Set default values without function call in function signature
        beta_x = beta_x if beta_x is not None else torch.tensor(0.0)
        alpha_x = alpha_x if alpha_x is not None else torch.tensor(0.0)
        emittance_x = emittance_x if emittance_x is not None else torch.tensor(0.0)
        beta_y = beta_y if beta_y is not None else torch.tensor(0.0)
        alpha_y = alpha_y if alpha_y is not None else torch.tensor(0.0)
        emittance_y = emittance_y if emittance_y is not None else torch.tensor(0.0)
        sigma_s = sigma_s if sigma_s is not None else torch.tensor(1e-6)
        sigma_p = sigma_p if sigma_p is not None else torch.tensor(1e-6)
        cor_s = cor_s if cor_s is not None else torch.tensor(0.0)
        energy = energy if energy is not None else torch.tensor(1e8)

        sigma_x = torch.sqrt(emittance_x * beta_x)
        sigma_xp = torch.sqrt(emittance_x * (1 + alpha_x**2) / beta_x)
        sigma_y = torch.sqrt(emittance_y * beta_y)
        sigma_yp = torch.sqrt(emittance_y * (1 + alpha_y**2) / beta_y)
        cor_x = -emittance_x * alpha_x
        cor_y = -emittance_y * alpha_y
        return cls.from_parameters(
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
            cor_s=cor_s,
            cor_x=cor_x,
            cor_y=cor_y,
            device=device,
        )

    @classmethod
    def from_ocelot(cls, parray, device: str = "auto") -> "ParameterBeam":
        mu = torch.ones(7)
        mu[:6] = torch.tensor(parray.rparticles.mean(axis=1), dtype=torch.float32)

        cov = torch.zeros(7, 7)
        cov[:6, :6] = torch.tensor(np.cov(parray.rparticles), dtype=torch.float32)

        energy = torch.tensor(1e9 * parray.E)

        return cls(mu=mu, cov=cov, energy=energy, device=device)

    @classmethod
    def from_astra(cls, path: str, **kwargs) -> "ParameterBeam":
        """Load an Astra particle distribution as a Cheetah Beam."""
        from cheetah.converters.astralavista import from_astrabeam

        particles, energy = from_astrabeam(path)
        mu = torch.ones(7)
        mu[:6] = torch.tensor(particles.mean(axis=0))

        cov = torch.zeros(7, 7)
        cov[:6, :6] = torch.tensor(np.cov(particles.transpose()))

        return cls(mu=mu, cov=cov, energy=torch.tensor(energy), **kwargs)

    def transformed_to(
        self,
        mu_x: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
    ) -> "ParameterBeam":
        """
        Create version of this beam that is transformed to new beam parameters.

        :param n: Number of particles to generate.
        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_xp: Center of the particle distribution on px in meters.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_yp: Center of the particle distribution on py in meters.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_xp: Sigma of the particle distribution in px direction in meters.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_yp: Sigma of the particle distribution in py direction in meters.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param energy: Energy of the beam in eV.
        """
        mu_x = mu_x if mu_x is not None else self.mu_x
        mu_xp = mu_xp if mu_xp is not None else self.mu_xp
        mu_y = mu_y if mu_y is not None else self.mu_y
        mu_yp = mu_yp if mu_yp is not None else self.mu_yp
        sigma_x = sigma_x if sigma_x is not None else self.sigma_x
        sigma_xp = sigma_xp if sigma_xp is not None else self.sigma_xp
        sigma_y = sigma_y if sigma_y is not None else self.sigma_y
        sigma_yp = sigma_yp if sigma_yp is not None else self.sigma_yp
        sigma_s = sigma_s if sigma_s is not None else self.sigma_s
        sigma_p = sigma_p if sigma_p is not None else self.sigma_p
        energy = energy if energy is not None else self.energy

        return self.__class__.from_parameters(
            mu_x=mu_x,
            mu_xp=mu_xp,
            mu_y=mu_y,
            mu_yp=mu_yp,
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
            device=self.device,
        )

    @property
    def mu_x(self) -> torch.Tensor:
        return self._mu[0]

    @property
    def sigma_x(self) -> torch.Tensor:
        return torch.sqrt(self._cov[0, 0])

    @property
    def mu_xp(self) -> torch.Tensor:
        return self._mu[1]

    @property
    def sigma_xp(self) -> torch.Tensor:
        return torch.sqrt(self._cov[1, 1])

    @property
    def mu_y(self) -> torch.Tensor:
        return self._mu[2]

    @property
    def sigma_y(self) -> torch.Tensor:
        return torch.sqrt(self._cov[2, 2])

    @property
    def mu_yp(self) -> torch.Tensor:
        return self._mu[3]

    @property
    def sigma_yp(self) -> torch.Tensor:
        return torch.sqrt(self._cov[3, 3])

    @property
    def mu_s(self) -> torch.Tensor:
        return self._mu[4]

    @property
    def sigma_s(self) -> torch.Tensor:
        return torch.sqrt(self._cov[4, 4])

    @property
    def mu_p(self) -> torch.Tensor:
        return self._mu[5]

    @property
    def sigma_p(self) -> torch.Tensor:
        return torch.sqrt(self._cov[5, 5])

    @property
    def sigma_xxp(self) -> torch.Tensor:
        return self._cov[0, 1]

    @property
    def sigma_yyp(self) -> torch.Tensor:
        return self._cov[2, 3]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(mu_x={repr(self.mu_x)},"
            f" mu_xp={repr(self.mu_xp)}, mu_y={repr(self.mu_y)},"
            f" mu_yp={repr(self.mu_yp)}, sigma_x={repr(self.sigma_x)},"
            f" sigma_xp={repr(self.sigma_xp)}, sigma_y={repr(self.sigma_y)},"
            f" sigma_yp={repr(self.sigma_yp)}, sigma_s={repr(self.sigma_s)},"
            f" sigma_p={repr(self.sigma_p)}, energy={repr(self.energy)})"
        )


class ParticleBeam(Beam):
    """
    Beam of charged particles, where each particle is simulated.

    :param particles: List of 7-dimensional particle vectors.
    :param energy: Energy of the beam in eV.
    :param device: Device to move the beam's particle array to. If set to `"auto"` a
        CUDA GPU is selected if available. The CPU is used otherwise.
    """

    def __init__(
        self, particles: torch.Tensor, energy: torch.Tensor, device: str = "auto"
    ) -> None:
        super().__init__()

        assert (
            len(particles) > 0 and particles.shape[1] == 7
        ), "Particle vectors must be 7-dimensional."

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

        self.particles = (
            particles.to(device) if isinstance(particles, torch.Tensor) else particles
        )

        self.energy = energy

    @classmethod
    def from_parameters(
        cls,
        num_particles: Optional[torch.Tensor] = None,
        mu_x: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        cor_x: Optional[torch.Tensor] = None,
        cor_y: Optional[torch.Tensor] = None,
        cor_s: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
        device: str = "auto",
    ) -> "ParticleBeam":
        """
        Generate Cheetah Beam of random particles.

        :param num_particles: Number of particles to generate.
        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_xp: Center of the particle distribution on px in meters.
        :param mu_yp: Center of the particle distribution on py in meters.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_xp: Sigma of the particle distribution in px direction in meters.
        :param sigma_yp: Sigma of the particle distribution in py direction in meters.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param cor_x: Correlation between x and xp.
        :param cor_y: Correlation between y and yp.
        :param cor_s: Correlation between s and p.
        :param energy: Energy of the beam in eV.
        :param device: Device to move the beam's particle array to. If set to `"auto"` a
            CUDA GPU is selected if available. The CPU is used otherwise.
        """
        # Set default values without function call in function signature
        num_particles = (
            num_particles if num_particles is not None else torch.tensor(100_000)
        )
        mu_x = mu_x if mu_x is not None else torch.tensor(0.0)
        mu_xp = mu_xp if mu_xp is not None else torch.tensor(0.0)
        mu_y = mu_y if mu_y is not None else torch.tensor(0.0)
        mu_yp = mu_yp if mu_yp is not None else torch.tensor(0.0)
        sigma_x = sigma_x if sigma_x is not None else torch.tensor(175e-9)
        sigma_xp = sigma_xp if sigma_xp is not None else torch.tensor(2e-7)
        sigma_y = sigma_y if sigma_y is not None else torch.tensor(175e-9)
        sigma_yp = sigma_yp if sigma_yp is not None else torch.tensor(2e-7)
        sigma_s = sigma_s if sigma_s is not None else torch.tensor(1e-6)
        sigma_p = sigma_p if sigma_p is not None else torch.tensor(1e-6)
        cor_x = cor_x if cor_x is not None else torch.tensor(0.0)
        cor_y = cor_y if cor_y is not None else torch.tensor(0.0)
        cor_s = cor_s if cor_s is not None else torch.tensor(0.0)
        energy = energy if energy is not None else torch.tensor(1e8)

        mean = torch.stack(
            [mu_x, mu_xp, mu_y, mu_yp, torch.tensor(0.0), torch.tensor(0.0)]
        )

        cov = torch.zeros(6, 6)
        cov[0, 0] = sigma_x**2
        cov[0, 1] = cor_x
        cov[1, 0] = cor_x
        cov[1, 1] = sigma_xp**2
        cov[2, 2] = sigma_y**2
        cov[2, 3] = cor_y
        cov[3, 2] = cor_y
        cov[3, 3] = sigma_yp**2
        cov[4, 4] = sigma_s**2
        cov[4, 5] = cor_s
        cov[5, 4] = cor_s
        cov[5, 5] = sigma_p**2

        particles = torch.ones((num_particles, 7), dtype=torch.float32)
        distribution = MultivariateNormal(mean, covariance_matrix=cov)
        particles[:, :6] = distribution.sample((num_particles,))

        return cls(particles, energy, device=device)

    @classmethod
    def from_twiss(
        cls,
        num_particles: Optional[torch.Tensor] = None,
        beta_x: Optional[torch.Tensor] = None,
        alpha_x: Optional[torch.Tensor] = None,
        emittance_x: Optional[torch.Tensor] = None,
        beta_y: Optional[torch.Tensor] = None,
        alpha_y: Optional[torch.Tensor] = None,
        emittance_y: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        cor_s: Optional[torch.Tensor] = None,
        device: str = "auto",
    ) -> "ParticleBeam":
        # Set default values without function call in function signature
        num_particles = (
            num_particles if num_particles is not None else torch.tensor(1_000_000)
        )
        beta_x = beta_x if beta_x is not None else torch.tensor(0.0)
        alpha_x = alpha_x if alpha_x is not None else torch.tensor(0.0)
        emittance_x = emittance_x if emittance_x is not None else torch.tensor(0.0)
        beta_y = beta_y if beta_y is not None else torch.tensor(0.0)
        alpha_y = alpha_y if alpha_y is not None else torch.tensor(0.0)
        emittance_y = emittance_y if emittance_y is not None else torch.tensor(0.0)
        energy = energy if energy is not None else torch.tensor(1e8)
        sigma_s = sigma_s if sigma_s is not None else torch.tensor(1e-6)
        sigma_p = sigma_p if sigma_p is not None else torch.tensor(1e-6)
        cor_s = cor_s if cor_s is not None else torch.tensor(0.0)

        sigma_x = torch.sqrt(beta_x * emittance_x)
        sigma_xp = torch.sqrt(emittance_x * (1 + alpha_x**2) / beta_x)
        sigma_y = torch.sqrt(beta_y * emittance_y)
        sigma_yp = torch.sqrt(emittance_y * (1 + alpha_y**2) / beta_y)
        cor_x = -emittance_x * alpha_x
        cor_y = -emittance_y * alpha_y

        return cls.from_parameters(
            num_particles=num_particles,
            mu_x=torch.tensor(0.0),
            mu_xp=torch.tensor(0.0),
            mu_y=torch.tensor(0.0),
            mu_yp=torch.tensor(0.0),
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
            cor_s=cor_s,
            cor_x=cor_x,
            cor_y=cor_y,
            device=device,
        )

    @classmethod
    def make_linspaced(
        cls,
        num_particles: Optional[torch.Tensor] = None,
        mu_x: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
        device: str = "auto",
    ) -> "ParticleBeam":
        """
        Generate Cheetah Beam of *n* linspaced particles.

        :param n: Number of particles to generate.
        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_xp: Center of the particle distribution on px in meters.
        :param mu_yp: Center of the particle distribution on py in meters.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_xp: Sigma of the particle distribution in px direction in meters.
        :param sigma_yp: Sigma of the particle distribution in py direction in meters.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param energy: Energy of the beam in eV.
        :param device: Device to move the beam's particle array to. If set to `"auto"` a
            CUDA GPU is selected if available. The CPU is used otherwise.
        """

        # Set default values without function call in function signature
        num_particles = num_particles if num_particles is not None else torch.tensor(10)
        mu_x = mu_x if mu_x is not None else torch.tensor(0.0)
        mu_xp = mu_xp if mu_xp is not None else torch.tensor(0.0)
        mu_y = mu_y if mu_y is not None else torch.tensor(0.0)
        mu_yp = mu_yp if mu_yp is not None else torch.tensor(0.0)
        sigma_x = sigma_x if sigma_x is not None else torch.tensor(175e-9)
        sigma_xp = sigma_xp if sigma_xp is not None else torch.tensor(2e-7)
        sigma_y = sigma_y if sigma_y is not None else torch.tensor(175e-9)
        sigma_yp = sigma_yp if sigma_yp is not None else torch.tensor(2e-7)
        sigma_s = sigma_s if sigma_s is not None else torch.tensor(0.0)
        sigma_p = sigma_p if sigma_p is not None else torch.tensor(0.0)
        energy = energy if energy is not None else torch.tensor(1e8)

        particles = torch.ones((num_particles, 7), dtype=torch.float32)

        particles[:, 0] = torch.linspace(
            mu_x - sigma_x, mu_x + sigma_x, num_particles, dtype=torch.float32
        )
        particles[:, 1] = torch.linspace(
            mu_xp - sigma_xp, mu_xp + sigma_xp, num_particles, dtype=torch.float32
        )
        particles[:, 2] = torch.linspace(
            mu_y - sigma_y, mu_y + sigma_y, num_particles, dtype=torch.float32
        )
        particles[:, 3] = torch.linspace(
            mu_yp - sigma_yp, mu_yp + sigma_yp, num_particles, dtype=torch.float32
        )
        particles[:, 4] = torch.linspace(
            -sigma_s, sigma_s, num_particles, dtype=torch.float32
        )
        particles[:, 5] = torch.linspace(
            -sigma_p, sigma_p, num_particles, dtype=torch.float32
        )

        return cls(particles, energy, device=device)

    @classmethod
    def from_ocelot(cls, parray, device: str = "auto") -> "ParticleBeam":
        """
        Convert an Ocelot ParticleArray `parray` to a Cheetah Beam.
        """
        num_particles = parray.rparticles.shape[1]
        particles = torch.ones((num_particles, 7))
        particles[:, :6] = torch.tensor(
            parray.rparticles.transpose(), dtype=torch.float32
        )

        return cls(
            particles=particles, energy=torch.tensor(1e9 * parray.E), device=device
        )

    @classmethod
    def from_astra(cls, path: str, **kwargs) -> "ParticleBeam":
        """Load an Astra particle distribution as a Cheetah Beam."""
        from cheetah.converters.astralavista import from_astrabeam

        particles, energy = from_astrabeam(path)
        particles_7d = torch.ones((particles.shape[0], 7))
        particles_7d[:, :6] = torch.from_numpy(particles)
        return cls(particles=particles_7d, energy=torch.tensor(energy), **kwargs)

    def transformed_to(
        self,
        mu_x: Optional[torch.Tensor] = None,
        mu_y: Optional[torch.Tensor] = None,
        mu_xp: Optional[torch.Tensor] = None,
        mu_yp: Optional[torch.Tensor] = None,
        sigma_x: Optional[torch.Tensor] = None,
        sigma_y: Optional[torch.Tensor] = None,
        sigma_xp: Optional[torch.Tensor] = None,
        sigma_yp: Optional[torch.Tensor] = None,
        sigma_s: Optional[torch.Tensor] = None,
        sigma_p: Optional[torch.Tensor] = None,
        energy: Optional[torch.Tensor] = None,
    ) -> "ParticleBeam":
        """
        Create version of this beam that is transformed to new beam parameters.

        :param n: Number of particles to generate.
        :param mu_x: Center of the particle distribution on x in meters.
        :param mu_y: Center of the particle distribution on y in meters.
        :param mu_xp: Center of the particle distribution on px in meters.
        :param mu_yp: Center of the particle distribution on py in meters.
        :param sigma_x: Sigma of the particle distribution in x direction in meters.
        :param sigma_y: Sigma of the particle distribution in y direction in meters.
        :param sigma_xp: Sigma of the particle distribution in px direction in meters.
        :param sigma_yp: Sigma of the particle distribution in py direction in meters.
        :param sigma_s: Sigma of the particle distribution in s direction in meters.
        :param sigma_p: Sigma of the particle distribution in p direction in meters.
        :param energy: Energy of the beam in eV.
        :param device: Device to move the beam's particle array to. If set to `"auto"` a
            CUDA GPU is selected if available. The CPU is used otherwise.
        """
        mu_x = mu_x if mu_x is not None else self.mu_x
        mu_y = mu_y if mu_y is not None else self.mu_y
        mu_xp = mu_xp if mu_xp is not None else self.mu_xp
        mu_yp = mu_yp if mu_yp is not None else self.mu_yp
        sigma_x = sigma_x if sigma_x is not None else self.sigma_x
        sigma_y = sigma_y if sigma_y is not None else self.sigma_y
        sigma_xp = sigma_xp if sigma_xp is not None else self.sigma_xp
        sigma_yp = sigma_yp if sigma_yp is not None else self.sigma_yp
        sigma_s = sigma_s if sigma_s is not None else self.sigma_s
        sigma_p = sigma_p if sigma_p is not None else self.sigma_p
        energy = energy if energy is not None else self.energy

        new_mu = torch.stack(
            [mu_x, mu_xp, mu_y, mu_yp, torch.tensor(0.0), torch.tensor(0.0)]
        )
        new_sigma = torch.stack(
            [sigma_x, sigma_xp, sigma_y, sigma_yp, sigma_s, sigma_p]
        )

        old_mu = torch.stack(
            [
                self.mu_x,
                self.mu_xp,
                self.mu_y,
                self.mu_yp,
                torch.tensor(0.0),
                torch.tensor(0.0),
            ]
        )
        old_sigma = torch.stack(
            [
                self.sigma_x,
                self.sigma_xp,
                self.sigma_y,
                self.sigma_yp,
                self.sigma_s,
                self.sigma_p,
            ]
        )

        phase_space = self.particles[:, :6]
        phase_space = (phase_space - old_mu) / old_sigma * new_sigma + new_mu

        particles = torch.ones_like(
            self.particles, dtype=torch.float32, device=self.device
        )
        particles[:, :6] = phase_space

        return self.__class__(particles=particles, energy=energy)

    def __len__(self) -> int:
        return int(self.num_particles)

    @property
    def num_particles(self) -> torch.Tensor:
        return len(self.particles)

    @property
    def xs(self) -> Optional[torch.Tensor]:
        return self.particles[:, 0] if self is not Beam.empty else None

    @xs.setter
    def xs(self, value: torch.Tensor) -> None:
        self.particles[:, 0] = value

    @property
    def mu_x(self) -> Optional[torch.Tensor]:
        return self.xs.mean() if self is not Beam.empty else None

    @property
    def sigma_x(self) -> Optional[torch.Tensor]:
        return self.xs.std() if self is not Beam.empty else None

    @property
    def xps(self) -> Optional[torch.Tensor]:
        return self.particles[:, 1] if self is not Beam.empty else None

    @xps.setter
    def xps(self, value: torch.Tensor) -> None:
        self.particles[:, 1] = value

    @property
    def mu_xp(self) -> Optional[torch.Tensor]:
        return self.xps.mean() if self is not Beam.empty else None

    @property
    def sigma_xp(self) -> Optional[torch.Tensor]:
        return self.xps.std() if self is not Beam.empty else None

    @property
    def ys(self) -> Optional[torch.Tensor]:
        return self.particles[:, 2] if self is not Beam.empty else None

    @ys.setter
    def ys(self, value: torch.Tensor) -> None:
        self.particles[:, 2] = value

    @property
    def mu_y(self) -> Optional[float]:
        return self.ys.mean() if self is not Beam.empty else None

    @property
    def sigma_y(self) -> Optional[torch.Tensor]:
        return self.ys.std() if self is not Beam.empty else None

    @property
    def yps(self) -> Optional[torch.Tensor]:
        return self.particles[:, 3] if self is not Beam.empty else None

    @yps.setter
    def yps(self, value: torch.Tensor) -> None:
        self.particles[:, 3] = value

    @property
    def mu_yp(self) -> Optional[torch.Tensor]:
        return self.yps.mean() if self is not Beam.empty else None

    @property
    def sigma_yp(self) -> Optional[torch.Tensor]:
        return self.yps.std() if self is not Beam.empty else None

    @property
    def ss(self) -> Optional[torch.Tensor]:
        return self.particles[:, 4] if self is not Beam.empty else None

    @ss.setter
    def ss(self, value: torch.Tensor) -> None:
        self.particles[:, 4] = value

    @property
    def mu_s(self) -> Optional[torch.Tensor]:
        return self.ss.mean() if self is not Beam.empty else None

    @property
    def sigma_s(self) -> Optional[torch.Tensor]:
        return self.ss.std() if self is not Beam.empty else None

    @property
    def ps(self) -> Optional[torch.Tensor]:
        return self.particles[:, 5] if self is not Beam.empty else None

    @ps.setter
    def ps(self, value: torch.Tensor) -> None:
        self.particles[:, 5] = value

    @property
    def mu_p(self) -> Optional[torch.Tensor]:
        return self.ps.mean() if self is not Beam.empty else None

    @property
    def sigma_p(self) -> Optional[torch.Tensor]:
        return self.ps.std() if self is not Beam.empty else None

    @property
    def sigma_xxp(self) -> torch.Tensor:
        return torch.mean((self.xs - self.mu_x) * (self.xps - self.mu_xp))

    @property
    def sigma_yyp(self) -> torch.Tensor:
        return torch.mean((self.ys - self.mu_y) * (self.yps - self.mu_yp))

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(n={repr(self.num_particles)},"
            f" mu_x={repr(self.mu_x)}, mu_xp={repr(self.mu_xp)},"
            f" mu_y={repr(self.mu_y)}, mu_yp={repr(self.mu_yp)},"
            f" sigma_x={repr(self.sigma_x)}, sigma_xp={repr(self.sigma_xp)},"
            f" sigma_y={repr(self.sigma_y)}, sigma_yp={repr(self.sigma_yp)},"
            f" sigma_s={repr(self.sigma_s)}, sigma_p={repr(self.sigma_p)},"
            f" energy={repr(self.energy)})"
        )



class DeviceError(Exception):
    """
    Used to create an exception, in case the device used for the beam
    and the elements are different.
    """

    def __init__(self):
        super().__init__(
            "Warning! The device used for calculating the elements is not the same, "
            "as the device used to calculate the Beam."
        )
